@@ -21,7 +21,7 @@
 (They are monkey-patched by other modules)
 
 """
-__all__ = ["ASTNode", "Expression", "Constructor", "IfExpr", "Quantee", "AQuantification",
+__all__ = ["ASTNode", "Expression", "Constructor", "AIfExpr", "Quantee", "AQuantification",
            "Operator", "AImplication", "AEquivalence", "ARImplication",
            "ADisjunction", "AConjunction", "AComparison", "ASumMinus",
            "AMultDiv", "APower", "AUnary", "AAggregate", "AppliedSymbol",
@@ -31,12 +31,13 @@ __all__ = ["ASTNode", "Expression", "Constructor", "IfExpr", "Quantee", "AQuanti
 import copy
 from collections import ChainMap
 from datetime import date
+from fractions import Fraction
+from re import findall
 from sys import intern
 from textx import get_location
 from typing import Optional, List, Tuple, Dict, Set, Any
-from z3 import BoolSort, IntSort, RealSort
 
-from .utils import unquote, OrderedSet, BOOL, INT, REAL, RESERVED_SYMBOLS, \
+from .utils import unquote, OrderedSet, BOOL, INT, REAL, DATE, CONCEPT, RESERVED_SYMBOLS, \
     IDPZ3Error, DEF_SEMANTICS, Semantics
 
 
@@ -95,6 +96,38 @@ class ASTNode(object):
         return self  # monkey-patched
 
 
+class Annotations(ASTNode):
+    def __init__(self, **kwargs):
+        annotations = kwargs.pop('annotations')
+
+        self.annotations = {}
+        for s in annotations:
+            p = s.split(':', 1)
+            if len(p) == 2:
+                try:
+                    # Do we have a Slider?
+                    # The format of p[1] is as follows:
+                    # (lower_sym, upper_sym): (lower_bound, upper_bound)
+                    pat = r"\(((.*?), (.*?))\)"
+                    arg = findall(pat, p[1])
+                    l_symb = arg[0][1]
+                    u_symb = arg[0][2]
+                    l_bound = arg[1][1]
+                    u_bound = arg[1][2]
+                    slider_arg = {'lower_symbol': l_symb,
+                                  'upper_symbol': u_symb,
+                                  'lower_bound': l_bound,
+                                  'upper_bound': u_bound}
+                    k, v = (p[0], slider_arg)
+                except:  # could not parse the slider data
+                    k, v = (p[0], p[1])
+            else:
+                k, v = ('reading', p[0])
+            self.check(k not in self.annotations,
+                       f"Duplicate annotation: [{k}: {v}]")
+            self.annotations[k] = v
+
+
 class Constructor(ASTNode):
     """Constructor declaration
 
@@ -111,8 +144,6 @@ class Constructor(ASTNode):
         has been applied to some arguments (e.g., is_rgb)
 
         symbol (Symbol): only for Symbol constructors
-
-        translated (DataTypeRef): the value in Z3
     """
 
     def __init__(self, **kwargs):
@@ -126,7 +157,6 @@ class Constructor(ASTNode):
         self.type = None
         self.symbol = None
         self.tester = None
-        self.translated: Any = None
 
     def __str__(self):
         return (self.name if not self.sorts else
@@ -153,7 +183,6 @@ class Accessor(ASTNode):
                 f"{self.accessor}: {self.type}" )
 
 
-
 class Expression(ASTNode):
     """The abstract class of AST nodes representing (sub-)expressions.
 
@@ -199,7 +228,7 @@ class Expression(ASTNode):
             Equivalence is computed in the context of the theory and structure.
 
         annotations (Dict[str, str]):
-            The set of annotations given by the expert in the IDP source code.
+            The set of annotations given by the expert in the IDP-Z3 program.
 
             ``annotations['reading']`` is the annotation
             giving the intended meaning of the expression (in English).
@@ -207,20 +236,17 @@ class Expression(ASTNode):
         original (Expression):
             The original expression, before propagation and simplification.
 
-        fresh_vars (Set(string)):
+        variables (Set(string)):
             The set of names of the variables in the expression.
 
         is_type_constraint_for (string):
             name of the symbol for which the expression is a type constraint
 
-        translated (Optional[z3 ast]):
-            The translation of the expression to Z3 (cache)
-
     """
-    __slots__ = ('sub_exprs', 'simpler', 'value', 'status', 'code',
-                 'annotations', 'original', 'str', 'fresh_vars', 'type',
-                 '_reified', 'is_type_constraint_for', 'co_constraint',
-                 'normal', 'questions', 'relevant')
+    __slots__ = ('sub_exprs', 'simpler', 'value', 'code',
+                 'annotations', 'original', 'str', 'variables', 'type',
+                 'is_type_constraint_for', 'co_constraint',
+                 'questions', 'relevant')
 
     def __init__(self):
         self.sub_exprs: List["Expression"]
@@ -228,21 +254,21 @@ class Expression(ASTNode):
         self.value: Optional["Expression"] = None
 
         self.code: str = intern(str(self))
-        self.annotations: Dict[str, str] = {'reading': self.code}
+        if not hasattr(self, 'annotations') or self.annotations == None:
+            self.annotations: Dict[str, str] = {'reading': self.code}
+        elif type(self.annotations) == Annotations:
+            self.annotations = self.annotations.annotations
         self.original: Expression = self
 
         self.str: str = self.code
-        self.fresh_vars: Optional[Set[str]] = None
+        self.variables: Optional[Set[str]] = None
         self.type: Optional[str] = None
-        self._reified: Optional["Expression"] = None
         self.is_type_constraint_for: Optional[str] = None
         self.co_constraint: Optional["Expression"] = None
 
         # attributes of the top node of a (co-)constraint
         self.questions: Optional[OrderedSet] = None
         self.relevant: Optional[bool] = None
-        self.block: Any = None  # deprecated
-        self.translated = None
 
     def copy(self):
         " create a deep copy (except for rigid terms and variables) "
@@ -250,7 +276,7 @@ class Expression(ASTNode):
             return self
         out = copy.copy(self)
         out.sub_exprs = [e.copy() for e in out.sub_exprs]
-        out.fresh_vars = copy.copy(out.fresh_vars)
+        out.variables = copy.copy(out.variables)
         out.value = None if out.value is None else out.value.copy()
         out.simpler = None if out.simpler is None else out.simpler.copy()
         out.co_constraint = (None if out.co_constraint is None
@@ -260,7 +286,7 @@ class Expression(ASTNode):
         return out
 
     def same_as(self, other):
-        if id(self) == id(other):
+        if self.str == other.str:
             return True
         if self.value is not None and self.value is not self:
             return self.value  .same_as(other)
@@ -307,7 +333,7 @@ class Expression(ASTNode):
         and AppliedSymbol interpreted in a structure
         co_constraints=False : ignore co_constraints
 
-        default implementation for UnappliedSymbol, IfExpr, AUnary, Variable,
+        default implementation for UnappliedSymbol, AIfExpr, AUnary, Variable,
         Number_constant, Brackets
         """
         for e in self.sub_exprs:
@@ -352,7 +378,7 @@ class Expression(ASTNode):
 
         `co_constraints` is an OrderedSet of Expression
         """
-        if self.co_constraint is not None:
+        if self.co_constraint is not None and self.co_constraint not in co_constraints:
             co_constraints.append(self.co_constraint)
             self.co_constraint.co_constraints(co_constraints)
         for e in self.sub_exprs:
@@ -403,14 +429,14 @@ class Expression(ASTNode):
     def instantiate(self,
                     e0: List["Expression"],
                     e1: List["Expression"],
-                    problem: "Problem"=None
+                    problem: "Theory"=None
                     ) -> "Expression":
         return self  # monkey-patched
 
     def instantiate1(self,
                     e0: "Expression",
                     e1: "Expression",
-                    problem: "Problem"=None
+                    problem: "Theory"=None
                     ) -> "Expression":
         return self  # monkey-patched
 
@@ -431,13 +457,13 @@ class Expression(ASTNode):
                    ):
         return  # monkey-patched
 
-    def translate(self):
+    def translate(self, problem: "Theory", vars={}):
         pass  # monkey-patched
 
-    def reified(self):
+    def reified(self, problem: "Theory"):
         pass  # monkey-patched
 
-    def translate1(self):
+    def translate1(self, problem: "Theory", vars={}):
         pass  # monkey-patched
 
     def as_set_condition(self) -> Tuple[Optional["AppliedSymbol"], Optional[bool], Optional["Enumeration"]]:
@@ -463,7 +489,7 @@ class Expression(ASTNode):
          are added to atoms containing recursive symbols.
 
         Arguments:
-            - level_symbols (dict[SymbolDeclaration, Symbol]): the level mapping
+            - level_symbols (Dict[SymbolDeclaration, Symbol]): the level mapping
               symbols as well as their corresponding recursive symbols
             - head (AppliedSymbol): head of the rule we are adding level mapping
               symbols to.
@@ -476,9 +502,9 @@ class Expression(ASTNode):
         Returns:
             Expression
         """
-        self.sub_exprs = [e.add_level_mapping(level_symbols, head, pos_justification, polarity)
-                          for e in self.sub_exprs]
-        return self
+        return (self.update_exprs((e.add_level_mapping(level_symbols, head, pos_justification, polarity)
+                                   for e in self.sub_exprs))
+                    .annotate1())  # update .variables
 
 
 class Symbol(Expression):
@@ -487,10 +513,8 @@ class Symbol(Expression):
     Attributes:
         name (string): name of the symbol
     """
-    TO = {'𝔹': BOOL, 'ℤ': INT, 'ℝ': REAL,
-          '`𝔹': '`'+BOOL, '`ℤ': '`'+INT, '`ℝ': '`'+REAL,}
-    FROM = {BOOL: '𝔹', INT: 'ℤ', REAL: 'ℝ',
-            '`'+BOOL: '`𝔹', '`'+INT: '`ℤ', '`'+REAL: '`ℝ',}
+    TO = {'Bool': BOOL, 'Int': INT, 'Real': REAL,
+          '`Bool': '`'+BOOL, '`Int': '`'+INT, '`Real': '`'+REAL,}
 
     def __init__(self, **kwargs):
         self.name = unquote(kwargs.pop('name'))
@@ -498,17 +522,78 @@ class Symbol(Expression):
         self.sub_exprs = []
         self.decl = None
         super().__init__()
-        self.fresh_vars = set()
+        self.variables = set()
         self.value = self
 
     def __str__(self):
-        return Symbol.FROM.get(self.name, self.name)
+        return self.name
 
     def __repr__(self):
         return str(self)
 
+    def has_element(self, term: Expression) -> Expression:
+        """Returns an expression that says whether `term` is in the type denoted by `self`.
 
-class IfExpr(Expression):
+        Args:
+            term (Expression): the argument to be checked
+
+        Returns:
+            Expression: whether `term` is in the type denoted by `self`.
+        """
+        return self.decl.check_bounds(term)
+
+
+class Subtype(Symbol):
+    """ASTNode representing `aType` or `Concept[aSignature]`, e.g., `Concept[T*T->Bool]`
+
+    Inherits from Symbol
+
+    Args:
+        name (Symbol): name of the concept
+
+        ins (List[Symbol], Optional): domain of the Concept signature, e.g., `[T, T]`
+
+        out (Symbol, Optional): range of the Concept signature, e.g., `Bool`
+    """
+
+    def __init__(self, **kwargs):
+        self.ins = kwargs.pop('ins', None)
+        self.out = kwargs.pop('out', None)
+        super().__init__(**kwargs)
+
+    def __str__(self):
+        return self.name + ("" if not self.out else
+                            f"[{'*'.join(str(s) for s in self.ins)}->{self.out}]")
+
+    def __eq__(self, other):
+        self.check(self.name != CONCEPT or self.out,
+                   f"`Concept` must be qualified with a type signature")
+        return (self.name == other.name and
+                (not self.out or (
+                    self.out == other.out and
+                    len(self.ins) == len(other.ins) and
+                    all(s==o for s, o in zip(self.ins, other.ins)))))
+
+    def range():
+        pass  # monkey-patched
+
+    def has_element(self, term: Expression) -> Expression:
+        """Returns an Expression that says whether `term` is in the type denoted by `self`.
+
+        Args:
+            term (Expression): the argument to be checked
+
+        Returns:
+            Expression: whether `term` `term` is in the type denoted by `self`.
+        """
+        if self.name == CONCEPT:
+            comparisons = [EQUALS([term, c]) for c in self.range()]
+            return OR(comparisons)
+        else:
+            return self.decl.check_bounds(term)
+
+
+class AIfExpr(Expression):
     PRECEDENCE = 10
     IF = 0
     THEN = 1
@@ -528,9 +613,9 @@ class IfExpr(Expression):
         return out.annotate1().simplify1()
 
     def __str1__(self):
-        return (f" if   {self.sub_exprs[IfExpr.IF  ].str}"
-                f" then {self.sub_exprs[IfExpr.THEN].str}"
-                f" else {self.sub_exprs[IfExpr.ELSE].str}")
+        return (f" if   {self.sub_exprs[AIfExpr.IF  ].str}"
+                f" then {self.sub_exprs[AIfExpr.THEN].str}"
+                f" else {self.sub_exprs[AIfExpr.ELSE].str}")
 
     def collect_nested_symbols(self, symbols, is_nested):
         return Expression.collect_nested_symbols(self, symbols, True)
@@ -538,19 +623,34 @@ class IfExpr(Expression):
 
 class Quantee(Expression):
     """represents the description of quantification, e.g., `x in T` or `(x,y) in P`
+    The `Concept` type may be qualified, e.g. `Concept[Color->Bool]`
 
     Attributes:
-        vars (List[List[Variable]): the (tuples of) variables being quantified
+        vars (List[List[Variable]]): the (tuples of) variables being quantified
 
-        sub_exprs (List[SymbolExpr], Optional): the type or predicate to quantify over
+        subtype (Subtype, Optional): a literal Subtype to quantify over, e.g., `Color` or `Concept[Color->Bool]`.
 
-        arity (int): the length of the tuple of variable
+        sort (SymbolExpr, Optional): a dereferencing expression, e.g.,. `$(i)`.
+
+        sub_exprs (List[SymbolExpr], Optional): the (unqualified) type or predicate to quantify over,
+        e.g., `[Color], [Concept] or [$(i)]`.
+
+        arity (int): the length of the tuple of variables
+
+        decl (SymbolDeclaration, Optional): the (unqualified) Declaration to quantify over, after resolution of `$(i)`.
+        e.g., the declaration of `Color`
     """
     def __init__(self, **kwargs):
         self.vars = kwargs.pop('vars')
-        sort = kwargs.pop('sort')
-        self.sub_exprs = [sort] if sort else []
+        self.subtype = kwargs.pop('subtype') if 'subtype' in kwargs else None
+        sort = kwargs.pop('sort') if 'sort' in kwargs else None
+        if self.subtype:
+            self.check(self.subtype.name == CONCEPT or self.subtype.out is None,
+                   f"Can't use signature after predicate {self.subtype.name}")
 
+        self.sub_exprs = ([sort] if sort else
+                          [self.subtype] if self.subtype else
+                          [])
         self.arity = None
         for i, v in enumerate(self.vars):
             if hasattr(v, 'vars'):  # varTuple
@@ -568,20 +668,38 @@ class Quantee(Expression):
 
     @classmethod
     def make(cls, var, sort):
-        if sort and type(sort) != SymbolExpr:
-            sort = SymbolExpr(eval='', s=sort).annotate1()
         out = (cls) (vars=[var], sort=sort)
         return out.annotate1()
 
     def __str1__(self):
+        signature = ("" if len(self.sub_exprs) <= 1 else
+                     f"[{','.join(t.str for t in self.sub_exprs[1:-1])}->{self.sub_exprs[-1]}]"
+        )
         return (f"{','.join(str(v) for vs in self.vars for v in vs)} "
-                f"∈ {self.sub_exprs[0] if self.sub_exprs else None}")
+                f"∈ {self.sub_exprs[0] if self.sub_exprs else None}"
+                f"{signature}")
 
 
 class AQuantification(Expression):
+    """ASTNode representing a quantified formula
+
+    Args:
+        annotations (Dict[str, str]):
+            The set of annotations given by the expert in the IDP-Z3 program.
+
+            ``annotations['reading']`` is the annotation
+            giving the intended meaning of the expression (in English).
+
+        q (str): either '∀' or '∃'
+
+        quantees (List[Quantee]): list of variable declarations
+
+        f (Expression): the formula being quantified
+    """
     PRECEDENCE = 20
 
     def __init__(self, **kwargs):
+        self.annotations = kwargs.pop('annotations')
         self.q = kwargs.pop('q')
         self.quantees = kwargs.pop('quantees')
         self.f = kwargs.pop('f')
@@ -602,9 +720,7 @@ class AQuantification(Expression):
     @classmethod
     def make(cls, q, quantees, f, annotations=None):
         "make and annotate a quantified formula"
-        out = cls(q=q, quantees=quantees, f=f)
-        if annotations:
-            out.annotations = annotations
+        out = cls(annotations=annotations, q=q, quantees=quantees, f=f)
         return out.annotate1()
 
     def __str1__(self):
@@ -633,6 +749,10 @@ class AQuantification(Expression):
             q.collect_symbols(symbols, co_constraints)
         return symbols
 
+def FORALL(qs, expr, annotations=None):
+    return AQuantification.make('∀', qs, expr, annotations)
+def EXISTS(qs, expr, annotations=None):
+    return AQuantification.make('∃', qs, expr, annotations)
 
 class Operator(Expression):
     PRECEDENCE = 0  # monkey-patched
@@ -669,7 +789,7 @@ class Operator(Expression):
             return operands[0]
         if isinstance(ops, str):
             ops = [ops] * (len(operands)-1)
-        out = (cls)(sub_exprs=operands, operator=ops)
+        out = (cls)(annotations=annotations, sub_exprs=operands, operator=ops)
         if annotations:
             out.annotations = annotations
         return out.annotate1().simplify1()
@@ -698,32 +818,39 @@ class AImplication(Operator):
     PRECEDENCE = 50
 
     def add_level_mapping(self, level_symbols, head, pos_justification, polarity):
-        self.sub_exprs = [self.sub_exprs[0].add_level_mapping(level_symbols, head, pos_justification, not polarity),
-                          self.sub_exprs[1].add_level_mapping(level_symbols, head, pos_justification, polarity)]
-        return self
+        sub_exprs = [self.sub_exprs[0].add_level_mapping(level_symbols, head, pos_justification, not polarity),
+                     self.sub_exprs[1].add_level_mapping(level_symbols, head, pos_justification, polarity)]
+        return self.update_exprs(sub_exprs).annotate1()
 
+def IMPLIES(exprs, annotations=None):
+    return AImplication.make('⇒', exprs, annotations)
 
 class AEquivalence(Operator):
     PRECEDENCE = 40
 
     # NOTE: also used to split rules into positive implication and negative implication. Please don't change.
     def split(self):
-        posimpl = AImplication.make('⇒', [self.sub_exprs[0], self.sub_exprs[1]])
-        negimpl = ARImplication.make('⇐', [self.sub_exprs[0].copy(), self.sub_exprs[1].copy()])
-        return AConjunction.make('∧', [posimpl, negimpl])
+        posimpl = IMPLIES([self.sub_exprs[0], self.sub_exprs[1]])
+        negimpl = RIMPLIES([self.sub_exprs[0].copy(), self.sub_exprs[1].copy()])
+        return AND([posimpl, negimpl])
 
     def split_equivalences(self):
         out = self.update_exprs(e.split_equivalences() for e in self.sub_exprs)
         return out.split()
 
+def EQUIV(exprs, annotations=None):
+    return AEquivalence.make('⇔', exprs, annotations)
+
 class ARImplication(Operator):
     PRECEDENCE = 30
 
     def add_level_mapping(self, level_symbols, head, pos_justification, polarity):
-        self.sub_exprs = [self.sub_exprs[0].add_level_mapping(level_symbols, head, pos_justification, polarity),
-                          self.sub_exprs[1].add_level_mapping(level_symbols, head, pos_justification, not polarity)]
-        return self
+        sub_exprs = [self.sub_exprs[0].add_level_mapping(level_symbols, head, pos_justification, polarity),
+                     self.sub_exprs[1].add_level_mapping(level_symbols, head, pos_justification, not polarity)]
+        return self.update_exprs(sub_exprs).annotate1()
 
+def RIMPLIES(exprs, annotations):
+    return ARImplication.make('⇐', exprs, annotations)
 
 class ADisjunction(Operator):
     PRECEDENCE = 60
@@ -733,15 +860,20 @@ class ADisjunction(Operator):
             return super().__str1__()
         return f"{self.sub_exprs[0].sub_exprs[0].code} in {{{self.enumerated}}}"
 
+def OR(exprs):
+    return ADisjunction.make('∨', exprs)
 
 class AConjunction(Operator):
     PRECEDENCE = 70
 
+def AND(exprs):
+    return AConjunction.make('∧', exprs)
 
 class AComparison(Operator):
     PRECEDENCE = 80
 
     def __init__(self, **kwargs):
+        self.annotations = kwargs.pop('annotations')
         super().__init__(**kwargs)
 
     def is_assignment(self):
@@ -753,6 +885,8 @@ class AComparison(Operator):
                         for e in self.sub_exprs[0].sub_exprs) \
                 and self.sub_exprs[1].value is not None
 
+def EQUALS(exprs):
+    return AComparison.make('=',exprs)
 
 class ASumMinus(Operator):
     PRECEDENCE = 90
@@ -790,41 +924,41 @@ class AUnary(Expression):
         return f"{self.operator}({self.sub_exprs[0].str})"
 
     def add_level_mapping(self, level_symbols, head, pos_justification, polarity):
-        self.sub_exprs = [e.add_level_mapping(level_symbols, head,
-                                              pos_justification,
-                                              not polarity
-                                              if self.operator == '¬' else polarity)
-                          for e in self.sub_exprs]
-        return self
+        sub_exprs = (e.add_level_mapping(level_symbols, head,
+                                         pos_justification,
+                                         not polarity
+                                         if self.operator == '¬' else polarity)
+                     for e in self.sub_exprs)
+        return self.update_exprs(sub_exprs).annotate1()
 
+def NOT(expr):
+    return AUnary.make('¬', expr)
 
 class AAggregate(Expression):
     PRECEDENCE = 130
-    CONDITION = 0
-    OUT = 1
 
     def __init__(self, **kwargs):
         self.aggtype = kwargs.pop('aggtype')
         self.quantees = kwargs.pop('quantees')
         self.f = kwargs.pop('f')
-        self.out = kwargs.pop('out')
 
-        self.sub_exprs = [self.f, self.out] if self.out else [self.f]  # later: expressions to be summed
-        self.using_if = False  # cannot test q_vars, because aggregate may not have quantee
+        self.aggtype = "#" if self.aggtype == "card" else self.aggtype
+        self.sub_exprs = [self.f]  # later: expressions to be summed
+        self.annotated = False  # cannot test q_vars, because aggregate may not have quantee
+        self.q = ''
         super().__init__()
 
-        if self.aggtype == "sum" and self.out is None:
-            raise Exception("Must have output variable for sum")
-        if self.aggtype != "sum" and self.out is not None:
-            raise Exception("Can't have output variable for  #")
 
     def __str1__(self):
-        if not self.using_if:
+        if not self.annotated:
             vars = "".join([f"{q}" for q in self.quantees])
-            output = f" : {self.sub_exprs[AAggregate.OUT].str}" if self.out else ""
-            out = (f"{self.aggtype}{{{vars} : "
-                   f"{self.sub_exprs[AAggregate.CONDITION].str}"
-                   f"{output}}}")
+            out = ((f"{self.aggtype}(lambda {vars} : "
+                    f"{self.sub_exprs[0].str}"
+                    f")" ) if self.aggtype != "#" else
+                   (f"{self.aggtype}{{{vars} : "
+                    f"{self.sub_exprs[0].str}"
+                    f"}}")
+            )
         else:
             out = (f"{self.aggtype}{{"
                    f"{','.join(e.str for e in self.sub_exprs)}"
@@ -866,6 +1000,7 @@ class AppliedSymbol(Expression):
     PRECEDENCE = 200
 
     def __init__(self, **kwargs):
+        self.annotations = kwargs.pop('annotations')
         self.symbol = kwargs.pop('symbol')
         self.sub_exprs = kwargs.pop('sub_exprs')
         if 'is_enumerated' in kwargs:
@@ -890,7 +1025,7 @@ class AppliedSymbol(Expression):
 
     @classmethod
     def make(cls, symbol, args, **kwargs):
-        out = cls(symbol=symbol, sub_exprs=args, **kwargs)
+        out = cls(annotations=None, symbol=symbol, sub_exprs=args, **kwargs)
         out.sub_exprs = args
         # annotate
         out.decl = symbol.decl
@@ -900,7 +1035,7 @@ class AppliedSymbol(Expression):
     def construct(cls, constructor, args):
         out= cls.make(Symbol(name=constructor.name), args)
         out.decl = constructor
-        out.fresh_vars = {}
+        out.variables = {}
         return out
 
     def __str1__(self):
@@ -948,6 +1083,10 @@ class AppliedSymbol(Expression):
             or any(e.has_decision() for e in self.sub_exprs)
 
     def type_inference(self):
+        if self.symbol.decl:
+            self.check(self.symbol.decl.arity == len(self.sub_exprs),
+                f"Incorrect number of arguments in {self}: "
+                f"should be {self.symbol.decl.arity}")
         try:
             out = {}
             for i, e in enumerate(self.sub_exprs):
@@ -968,11 +1107,9 @@ class AppliedSymbol(Expression):
         return (self.in_enumeration or self.is_enumerated
                 or not all(e.value is not None for e in self.sub_exprs))
 
-    def reified(self):
-        if self._reified is None:
-            self._reified = ( super().reified() if self.is_reified() else
-                 self.translate() )
-        return self._reified
+    def reified(self, problem: "Theory"):
+        return ( super().reified(problem) if self.is_reified() else
+                 self.translate(problem) )
 
     def generate_constructors(self, constructors: dict):
         symbol = self.symbol.sub_exprs[0]
@@ -1001,9 +1138,9 @@ class AppliedSymbol(Expression):
                 AppliedSymbol.make(level_symbols[self.symbol.decl], self.sub_exprs)
             ])
             if polarity:
-                return AConjunction.make('∧', [comp, self])
+                return AND([comp, self])
             else:
-                return ADisjunction.make('∨', [comp, self])
+                return OR([comp, self])
 
 
 class SymbolExpr(Expression):
@@ -1037,7 +1174,6 @@ class UnappliedSymbol(Expression):
 
         self.sub_exprs = []
         self.decl = None
-        self.translated = None
         self.is_enumerated = None
         self.is_enumeration = None
         self.in_enumeration = None
@@ -1049,7 +1185,7 @@ class UnappliedSymbol(Expression):
         """
         out = (cls)(s=Symbol(name=constructor.name))
         out.decl = constructor
-        out.fresh_vars = {}
+        out.variables = {}
         return out
 
     def __str1__(self): return self.name
@@ -1065,6 +1201,11 @@ FALSE = UnappliedSymbol.construct(FALSEC)
 
 class Variable(Expression):
     """AST node for a variable in a quantification or aggregate
+
+    Args:
+        name (str): name of the variable
+
+        sort (Optional[Symbol]): sort of the variable, if known
     """
     PRECEDENCE = 200
 
@@ -1072,18 +1213,20 @@ class Variable(Expression):
         self.name = kwargs.pop('name')
         sort = kwargs.pop('sort') if 'sort' in kwargs else None
         self.sort = sort
+        assert sort is None or isinstance(sort, Subtype) or isinstance(sort, Symbol)
 
         super().__init__()
 
         self.type = sort.decl.name if sort and sort.decl else ''
         self.sub_exprs = []
-        self.translated = None
-        self.fresh_vars = set([self.name])
+        self.variables = set([self.name])
 
     def __str1__(self): return self.name
 
     def copy(self): return self
 
+    def annotate1(self): return self
+
 
 class Number(Expression):
     PRECEDENCE = 200
@@ -1094,16 +1237,34 @@ class Number(Expression):
         super().__init__()
 
         self.sub_exprs = []
-        self.fresh_vars = set()
+        self.variables = set()
         self.value = self
 
-        self.translated = None
-        self.translate()  # also sets self.type
+        ops = self.number.split("/")
+        if len(ops) == 2:  # possible with str_to_IDP on Z3 value
+            self.py_value = Fraction(self.number)
+            self.type = REAL
+        elif '.' in self.number:
+            v = (self.number if not self.number.endswith('?') else
+                 self.number[:-1])
+            if "e" in v:
+                self.py_value = float(eval(v))
+            else:
+                self.py_value = Fraction(v)
+            self.type = REAL
+        else:
+            self.py_value = int(self.number)
+            self.type = INT
 
     def __str__(self): return self.number
 
     def is_reified(self): return False
 
+    def real(self):
+        """converts the INT number to REAL"""
+        self.check(self.type in [INT, REAL], f"Can't convert {self} to {REAL}")
+        return Number(number=str(float(self.py_value)))
+
 
 ZERO = Number(number='0')
 ONE = Number(number='1')
@@ -1120,11 +1281,11 @@ class Date(Expression):
         super().__init__()
 
         self.sub_exprs = []
-        self.fresh_vars = set()
+        self.variables = set()
         self.value = self
 
-        self.translated = None
-        self.translate()  # also sets self.type
+        self.py_value = self.date.toordinal()
+        self.type = DATE
 
     def __str__(self): return f"#{self.date.isoformat()}"
 
@@ -1136,16 +1297,13 @@ class Brackets(Expression):
 
     def __init__(self, **kwargs):
         self.f = kwargs.pop('f')
-        annotations = kwargs.pop('annotations')
+        self.annotations = kwargs.pop('annotations')
+        if not self.annotations:
+            self.annotations = {'reading': self.f.annotations['reading']}
         self.sub_exprs = [self.f]
 
         super().__init__()
-        if type(annotations) == dict:
-            self.annotations = annotations
-        elif annotations is None:
-            self.annotations = None
-        else:  # Annotations instance
-            self.annotations = annotations.annotations
+
 
     # don't @use_value, to have parenthesis
     def __str__(self): return f"({self.sub_exprs[0].str})"
