@@ -19,15 +19,14 @@
 
 """
 
-Methods to interpret a theory in a data structure
+Methods to ground / interpret a theory in a data structure
 
-* substitute a constant by its value in an expression
-* replace symbols interpreted in a structure by their interpretation
 * expand quantifiers
+* replace symbols interpreted in the structure by their interpretation
 
 This module also includes methods to:
 
-* substitute an node by another in an AST tree
+* substitute a node by another in an AST tree
 * instantiate an expresion, i.e. replace a variable by a value
 
 This module monkey-patches the ASTNode class and sub-classes.
@@ -40,22 +39,23 @@ import copy
 from itertools import product
 
 from .Assignments import Status as S
-from .Parse import (Extern, TypeDeclaration,
-                    SymbolDeclaration, Symbol, Rule, SymbolInterpretation,
+from .Parse import (Import, TypeDeclaration,
+                    SymbolDeclaration, Symbol, SymbolInterpretation,
                     FunctionEnum, Enumeration, Tuple, ConstructedFrom,
                     Definition)
-from .Expression import (SymbolExpr, Expression, Constructor, AQuantification,
-                    AImplication, AConjunction, ARImplication, AAggregate,
-                    AComparison, AUnary, AppliedSymbol, UnappliedSymbol,
-                    Variable, TRUE, AEquivalence)
-from .utils import BOOL, RESERVED_SYMBOLS, SYMBOL, OrderedSet, DEFAULT
+from .Expression import (AIfExpr, SymbolExpr, Expression, Constructor,
+                    AQuantification, Subtype, FORALL, IMPLIES, AND, AAggregate,
+                    NOT, AppliedSymbol, UnappliedSymbol,
+                    Variable, TRUE, Number)
+from .utils import (BOOL, RESERVED_SYMBOLS, CONCEPT, OrderedSet, DEFAULT,
+                    GOAL_SYMBOL, EXPAND)
 
 
-# class Extern  ###########################################################
+# class Import  ###########################################################
 
 def interpret(self, problem):
     pass
-Extern.interpret = interpret
+Import.interpret = interpret
 
 
 # class TypeDeclaration  ###########################################################
@@ -65,7 +65,7 @@ def interpret(self, problem):
         problem.interpretations[self.name].interpret(problem)
     if self.interpretation:
         self.constructors = self.interpretation.enumeration.constructors
-    self.translate()
+    self.translate(problem)
     if self.constructors:
         self.range = sum([c.interpret(problem).range for c in self.constructors], [])
     elif self.interpretation.enumeration:  # range declaration
@@ -77,17 +77,18 @@ TypeDeclaration.interpret = interpret
 # class SymbolDeclaration  ###########################################################
 
 def interpret(self, problem):
-    self.domain = list(product(*[s.decl.range for s in self.sorts]))
-    self.range = self.out.decl.range
+    assert all(isinstance(s, Subtype) for s in self.sorts), 'internal error'
+    self.in_domain = list(product(*[s.range() for s in self.sorts]))
+    self.range = self.out.range()
 
     # create instances
     if self.name not in RESERVED_SYMBOLS:
         self.instances = {}
-        for arg in self.domain:
+        for arg in self.in_domain:
             expr = AppliedSymbol.make(Symbol(name=self.name), arg)
             expr.annotate(self.voc, {})
             self.instances[expr.code] = expr
-            problem.assignments.assert__(expr, None, S.UNKNOWN, False)
+            problem.assignments.assert__(expr, None, S.UNKNOWN)
 
     # add type constraints to problem.constraints
     if self.out.decl.name != BOOL and self.name not in RESERVED_SYMBOLS:
@@ -107,7 +108,7 @@ def interpret(self, problem):
     """updates problem.def_constraints, by expanding the definitions
 
     Args:
-        problem (Problem):
+        problem (Theory):
             containts the enumerations for the expansion; is updated with the expanded definitions
     """
     self.cache = {}  # reset the cache
@@ -121,19 +122,18 @@ def add_def_constraints(self, instantiables, problem, result):
     The `instantiables` (of the definition) are expanded in `problem`.
 
     Args:
-        instantiables (dict[SymbolDeclaration, list[Expression]]):
+        instantiables (Dict[SymbolDeclaration, list[Expression]]):
             the constraints without the quantification
 
-        problem (Problem):
+        problem (Theory):
             contains the structure for the expansion/interpretation of the constraints
 
-        result (dict[SymbolDeclaration, Definition, list[Expression]]):
+        result (Dict[SymbolDeclaration, Definition, list[Expression]]):
             a mapping from (Symbol, Definition) to the list of constraints
     """
     for decl, bodies in instantiables.items():
         quantees = self.canonicals[decl][0].quantees  # take quantee from 1st renamed rule
-        expr = [AQuantification.make('∀', quantees, e, e.annotations)
-                .interpret(problem)
+        expr = [FORALL(quantees, e, e.annotations).interpret(problem)
                 for e in bodies]
         result[decl, self] = expr
 Definition.add_def_constraints = add_def_constraints
@@ -142,34 +142,47 @@ Definition.add_def_constraints = add_def_constraints
 # class SymbolInterpretation  ###########################################################
 
 def interpret(self, problem):
-    status = S.STRUCTURE if self.block.name != DEFAULT else S.GIVEN
+    status = S.DEFAULT if self.block.name == DEFAULT else S.STRUCTURE
     if self.is_type_enumeration:
         self.enumeration.interpret(problem)
         self.symbol.decl.interpretation = self
-    else: # update problem.assignments with data from enumeration
+    elif not self.name in [GOAL_SYMBOL, EXPAND]:
+        # update problem.assignments with data from enumeration
         for t in self.enumeration.tuples:
+
             if type(self.enumeration) == FunctionEnum:
                 args, value = t.args[:-1], t.args[-1]
+                self.check(self.symbol.decl.has_in_range(value).same_as(TRUE),
+                           f"{value} is not in the range of {self.symbol.name}")
             else:
                 args, value = t.args, TRUE
+
+            # check that the arguments are in the domain
+            a = (str(args) if 1<len(args) else
+                    str(args[0]) if len(args)==1 else
+                    "()")
+            self.check(len(self.symbol.decl.sorts) == len(args),
+                f"Incorrect arity of {a} for {self.name}")
+            self.check(self.symbol.decl.has_in_domain(args).same_as(TRUE),
+                        f"{a} is not in the domain of {self.symbol.name}")
+
             expr = AppliedSymbol.make(self.symbol, args)
             self.check(expr.code not in problem.assignments
                 or problem.assignments[expr.code].status == S.UNKNOWN,
                 f"Duplicate entry in structure for '{self.name}': {str(expr)}")
-            e = problem.assignments.assert__(expr, value, status, False)
-            if (status == S.GIVEN  # for proper display in IC
+            e = problem.assignments.assert__(expr, value, status)
+            if (status == S.DEFAULT  # for proper display in IC
                 and type(self.enumeration) == FunctionEnum):
-                problem.assignments.assert__(e.formula(), TRUE, status, False)
+                problem.assignments.assert__(e.formula(), TRUE, status)
         if self.default is not None:
             for code, expr in self.symbol.decl.instances.items():
                 if (code not in problem.assignments
                     or problem.assignments[code].status != status):
-                    e = problem.assignments.assert__(expr, self.default, status,
-                                                False)
-                    if (status == S.GIVEN  # for proper display in IC
+                    e = problem.assignments.assert__(expr, self.default, status)
+                    if (status == S.DEFAULT  # for proper display in IC
+                        and type(self.enumeration) == FunctionEnum
                         and self.default.type != BOOL):
-                        problem.assignments.assert__(e.formula(), TRUE, status,
-                                                    False)
+                        problem.assignments.assert__(e.formula(), TRUE, status)
 SymbolInterpretation.interpret = interpret
 
 
@@ -198,7 +211,7 @@ def interpret(self, problem):
         self.range = [UnappliedSymbol.construct(self)]
     else:
         self.range = [AppliedSymbol.construct(self, e)
-                      for e in product(*[s.decl.out.decl.range for s in self.sorts])]
+                      for e in product(*[s.decl.out.range() for s in self.sorts])]
     return self
 Constructor.interpret = interpret
 
@@ -212,7 +225,7 @@ def interpret(self, problem) -> Expression:
     - instantiate definitions
 
     Args:
-        problem (Problem): the Problem to apply
+        problem (Theory): the Theory to apply
 
     Returns:
         Expression: the resulting expression
@@ -255,17 +268,17 @@ def instantiate(self, e0, e1, problem=None):
     """Recursively substitute Variable in e0 by e1 in a copy of self.
 
     Interpret appliedSymbols immediately if grounded (and not occurring in head of definition).
-    Update fresh_vars.
+    Update .variables.
     """
     assert all(type(e) == Variable for e in e0), \
            f"Internal error: instantiate {e0}"
     if self.value:
         return self
-    if problem and all(e.name not in self.fresh_vars for e in e0):
+    if problem and all(e.name not in self.variables for e in e0):
         return self.interpret(problem)
     out = copy.copy(self)  # shallow copy !
     out.annotations = copy.copy(out.annotations)
-    out.fresh_vars = copy.copy(out.fresh_vars)
+    out.variables = copy.copy(out.variables)
     return out.instantiate1(e0, e1, problem)
 Expression.instantiate = instantiate
 
@@ -273,7 +286,7 @@ def instantiate1(self, e0, e1, problem=None):
     """Recursively substitute Variable in e0 by e1 in self.
 
     Interpret appliedSymbols immediately if grounded (and not occurring in head of definition).
-    Update fresh_vars.
+    Update .variables.
     """
     # instantiate expressions, with simplification
     out = self.update_exprs(e.instantiate(e0, e1, problem)
@@ -282,11 +295,13 @@ def instantiate1(self, e0, e1, problem=None):
     if out.value is not None:  # replace by new value
         out = out.value
     else:
+        self.check(len(e0) == len(e1),
+                   f"Incorrect arity: {e0}, {e1}")
         for o, n in zip(e0, e1):
-            if o.name in out.fresh_vars:
-                out.fresh_vars.discard(o.name)
+            if o.name in out.variables:
+                out.variables.discard(o.name)
                 if type(n) == Variable:
-                    out.fresh_vars.add(n.name)
+                    out.variables.add(n.name)
             out.code = str(out)
     out.annotations['reading'] = out.code
     return out
@@ -300,13 +315,30 @@ def instantiate(self, e0, e1, problem=None):
 Symbol.instantiate = instantiate
 
 
+# class Subtype ###########################################################
+
+def range(self):
+    range = [t for t in self.decl.range]
+    if self.out:  # x in Concept[T->T]
+        range = [v for v in range
+                 if v.decl.symbol.decl.arity == len(self.ins)
+                 and isinstance(v.decl.symbol.decl, SymbolDeclaration)
+                 and v.decl.symbol.decl.out.name == self.out.name
+                 and len(v.decl.symbol.decl.sorts) == len(self.ins)
+                 and all(s == q
+                         for s, q in zip(v.decl.symbol.decl.sorts,
+                                         self.ins))]
+    return range
+Subtype.range = range
+
+
 # Class AQuantification  ######################################################
 
 def interpret(self, problem):
     """apply information in the problem and its vocabulary
 
     Args:
-        problem (Problem): the problem to be applied
+        problem (Theory): the problem to be applied
 
     Returns:
         Expression: the expanded quantifier expression
@@ -328,27 +360,33 @@ def interpret(self, problem):
             q.sub_exprs = [inferred[var.name]]
 
     forms = self.sub_exprs
+    assert len(forms)==1
     new_quantees = []
     for q in self.quantees:
-        self.check(q.sub_exprs[0].decl.out.type == BOOL,
-                    f"{q.sub_exprs[0]} is not a type or predicate")
-        if not q.sub_exprs[0].decl.range:
+        domain = q.sub_exprs[0]
+        self.check(domain.decl.out.type == BOOL,
+                    f"{domain} is not a type or predicate")
+        if not domain.decl.range:
             new_quantees.append(q)
         else:
-            if q.sub_exprs[0].code in problem.interpretations:
-                enumeration = problem.interpretations[q.sub_exprs[0].code].enumeration
+            if domain.code in problem.interpretations:
+                enumeration = problem.interpretations[domain.code].enumeration
                 range = [t.args for t in enumeration.tuples.values()]
                 guard = None
-            elif type(q.sub_exprs[0].decl) == SymbolDeclaration:
-                range = q.sub_exprs[0].decl.domain
-                guard = q.sub_exprs[0]
-            else:  # type declaration
-                range = [[t] for t in q.sub_exprs[0].decl.range] #TODO1 decl.enumeration.tuples
+            elif type(domain.decl) == SymbolDeclaration:
+                range = domain.decl.in_domain
+                guard = domain
+            elif isinstance(domain, Subtype):
+                range = [[t] for t in domain.range()]
+                guard = None
+            else:  # SymbolExpr (e.g. $(`Color))
+                range = [[t] for t in domain.decl.range] #TODO1 decl.enumeration.tuples
                 guard = None
 
+
             for vars in q.vars:
-                self.check(q.sub_exprs[0].decl.arity == len(vars),
-                            f"Incorrect arity of {q.sub_exprs[0]}")
+                self.check(domain.decl.arity == len(vars),
+                            f"Incorrect arity of {domain}")
                 out = []
                 for f in forms:
                     for val in range:
@@ -356,9 +394,18 @@ def interpret(self, problem):
                         if guard:  # adds `guard(val) =>` in front of expression
                             applied = AppliedSymbol.make(guard, val)
                             if self.q == '∀':
-                                new_f = AImplication.make('⇒', [applied, new_f])
-                            else:
-                                new_f = AConjunction.make('∧', [applied, new_f])
+                                new_f = IMPLIES([applied, new_f])
+                            elif self.q == '∃':
+                                new_f = AND([applied, new_f])
+                            else:  # aggregate
+                                if isinstance(new_f, AIfExpr):  # cardinality
+                                    # if a then b else 0 -> if (applied & a) then b else 0
+                                    arg1 = AND([applied,
+                                                        new_f.sub_exprs[0]])
+                                    new_f = AIfExpr.make(arg1, new_f.sub_exprs[1],
+                                                        new_f.sub_exprs[2])
+                                else:  # sum
+                                    new_f = AIfExpr.make(applied, new_f, Number(number="0"))
                         out.append(new_f)
                 forms = out
 
@@ -370,11 +417,11 @@ AQuantification.interpret = interpret
 
 
 def instantiate1(self, e0, e1, problem=None):
-    out = Expression.instantiate1(self, e0, e1, problem)  # updates fresh_vars
+    out = Expression.instantiate1(self, e0, e1, problem)  # updates .variables
     for q in self.quantees: # for !x in $(output_domain(s,1))
         if q.sub_exprs:
             q.sub_exprs[0] = q.sub_exprs[0].instantiate(e0, e1, problem)
-    if problem and not self.fresh_vars:  # expand nested quantifier if no variables left
+    if problem and not self.variables:  # expand nested quantifier if no variables left
         out = out.interpret(problem)
     return out
 AQuantification.instantiate1 = instantiate1
@@ -383,7 +430,7 @@ AQuantification.instantiate1 = instantiate1
 # Class AAggregate  ######################################################
 
 def interpret(self, problem):
-    assert self.using_if, f"Internal error in interpret"
+    assert self.annotated, f"Internal error in interpret"
     return AQuantification.interpret(self, problem)
 AAggregate.interpret = interpret
 
@@ -395,7 +442,7 @@ AAggregate.instantiate1 = instantiate1  # from AQuantification
 def interpret(self, problem):
     self.symbol = self.symbol.interpret(problem)
     sub_exprs = [e.interpret(problem) for e in self.sub_exprs]
-    simpler, co_constraint = None, None
+    value, simpler, co_constraint = None, None, None
     if self.decl:
         if self.is_enumerated:
             assert self.decl.type != BOOL, \
@@ -407,33 +454,36 @@ def interpret(self, problem):
                 else:
                     simpler = interpretation.enumeration.contains(sub_exprs, True)
                 if 'not' in self.is_enumerated:
-                    simpler = AUnary.make('¬', simpler)
+                    simpler = NOT(simpler)
                 simpler.annotations = self.annotations
         elif self.in_enumeration:
             # re-create original Applied Symbol
             core = AppliedSymbol.make(self.symbol, sub_exprs).copy()
             simpler = self.in_enumeration.contains([core], False)
             if 'not' in self.is_enumeration:
-                simpler = AUnary.make('¬', simpler)
+                simpler = NOT(simpler)
             simpler.annotations = self.annotations
         elif (self.decl.name in problem.interpretations
-            and any(s.decl.name == SYMBOL for s in self.decl.sorts)
+            # and any(s.decl.name == CONCEPT for s in self.decl.sorts)
             and all(a.value is not None for a in sub_exprs)):
             # apply enumeration of predicate over symbols to allow simplification
             # do not do it otherwise, for performance reasons
-            f = problem.interpretations[self.decl.name].interpret_application
-            simpler = f(problem, 0, self, sub_exprs)
-        if (not self.in_head and not self.fresh_vars):
+            interpretation = problem.interpretations[self.decl.name]
+            if interpretation.block.name != DEFAULT:
+                f = interpretation.interpret_application
+                value = f(problem, 0, self, sub_exprs)
+        if (not self.in_head and not self.variables):
             inst = [defin.instantiate_definition(self.decl, sub_exprs, problem)
                               for defin in problem.definitions]
             inst = [x for x in inst if x]
-            if len(inst) == 1:
-                co_constraint = inst[0]
-            elif len(inst) > 1:
-                co_constraint = AConjunction.make('∧', inst)
-    out = self._change(sub_exprs=sub_exprs, simpler=simpler,
-                       co_constraint=co_constraint)
-    return out
+            if inst:
+                co_constraint = AND(inst)
+        out = (value if value else
+               self._change(sub_exprs=sub_exprs, simpler=simpler,
+                        co_constraint=co_constraint))
+        return out
+    else:
+        return self
 AppliedSymbol.interpret = interpret
 
 
@@ -464,15 +514,21 @@ def substitute(self, e0, e1, assignments, tag=None):
 AppliedSymbol .substitute = substitute
 
 def instantiate1(self, e0, e1, problem=None):
-    out = Expression.instantiate1(self, e0, e1, problem)  # update fresh_vars
+    out = Expression.instantiate1(self, e0, e1, problem)  # update .variables
     if type(out) == AppliedSymbol:  # might be a number after instantiation
         if type(out.symbol) == SymbolExpr and out.symbol.is_intentional():  # $(x)()
             out.symbol = out.symbol.instantiate(e0, e1, problem)
             if type(out.symbol) == Symbol:  # found $(x)
                 self.check(len(out.sub_exprs) == len(out.symbol.decl.sorts),
                             f"Incorrect arity for {out.code}")
-                out = AppliedSymbol.make(out.symbol, out.sub_exprs)
-        if problem and not self.fresh_vars:
+                kwargs = ({'is_enumerated': out.is_enumerated} if out.is_enumerated else
+                          {'in_enumeration': out.in_enumeration} if out.in_enumeration else
+                          {})
+                out = AppliedSymbol.make(out.symbol, out.sub_exprs, **kwargs)
+                out.original = self
+        if out.co_constraint is not None:
+            out.co_constraint.instantiate(e0, e1, problem)
+        if problem and not self.variables:
             return out.interpret(problem)
     return out
 AppliedSymbol .instantiate1 = instantiate1
@@ -494,6 +550,8 @@ Variable.substitute = substitute
 def instantiate1(self, e0, e1, problem=None):
     if self.sort:
         self.sort = self.sort.instantiate(e0, e1, problem)
+    self.check(len(e0) == len(e1),
+               f"Incorrect arity: {e0}, {e1}")
     for o, n in zip(e0, e1):
         if self.code == o.code:
             return n
