@@ -20,9 +20,9 @@
 Classes to parse an IDP-Z3 theory.
 
 """
-__all__ = ["IDP", "Vocabulary", "Annotations", "Extern",
+__all__ = ["IDP", "Vocabulary", "Annotations", "Import",
            "TypeDeclaration",
-           "SymbolDeclaration", "Symbol", "Theory", "Definition",
+           "SymbolDeclaration", "Symbol", "TheoryBlock", "Definition",
            "Rule", "Structure", "Enumeration", "Tuple",
            "Display", "Procedure", ]
 
@@ -31,29 +31,42 @@ from datetime import date
 from enum import Enum
 from itertools import groupby
 from os import path
-from re import match, findall
+from re import match
 from sys import intern
 from textx import metamodel_from_file
 from typing import Dict, List, Union, Optional
 
 
 from .Assignments import Assignments
-from .Expression import (ASTNode, Constructor, Accessor, Symbol, SymbolExpr,
-                         IfExpr, AQuantification, Quantee,
+from .Expression import (Annotations, ASTNode, Constructor, Accessor, Symbol, SymbolExpr,
+                         Expression, AIfExpr, AQuantification, Subtype, Quantee,
                          ARImplication, AEquivalence,
                          AImplication, ADisjunction, AConjunction,
                          AComparison, ASumMinus, AMultDiv, APower, AUnary,
                          AAggregate, AppliedSymbol, UnappliedSymbol,
                          Number, Brackets, Date,
-                         Variable, TRUEC, FALSEC, TRUE, FALSE)
-from .utils import (OrderedSet, NEWL, BOOL, INT, REAL, DATE, SYMBOL,
-                    RELEVANT, ABS, ARITY, INPUT_DOMAIN, OUTPUT_DOMAIN, IDPZ3Error,
+                         Variable, TRUEC, FALSEC, TRUE, FALSE, EQUALS, AND, OR, EQUIV)
+from .utils import (RESERVED_SYMBOLS, OrderedSet, NEWL, BOOL, INT, REAL, DATE, CONCEPT,
+                    GOAL_SYMBOL, EXPAND, RELEVANT, ABS, IDPZ3Error,
                     CO_CONSTR_RECURSION_DEPTH, MAX_QUANTIFIER_EXPANSION)
 
 
 def str_to_IDP(atom, val_string):
+    """cast a string value for 'atom into an Expr object, or None
+
+    used to convert Z3 models or json data from GUI
+
+    Args:
+        atom (Expr): the atom whose value must be converted
+        val_string (str): the string representation of the value
+
+    Returns:
+        Expr?: the value cast as Expr, or None if unknown
+    """
     assert atom.type, "Internal error"
-    if atom.type == BOOL:
+    if val_string == str(atom) or val_string+"()" == str(atom):
+        out = None  # Z3 means the value is unknown
+    elif atom.type == BOOL:
         if val_string not in ['True', 'False', 'true', 'false']:
             raise IDPZ3Error(
                 f"{atom.annotations['reading']} has wrong value: {val_string}")
@@ -81,12 +94,8 @@ def str_to_IDP(atom, val_string):
                 for a in args]
 
         out = AppliedSymbol.construct(constructor, args)
-    else:
-        try:
-            # could be a fraction
-            out = Number(number=str(eval(val_string.replace('?', ''))))
-        except:
-            out = None  # when z3 model has unknown value
+    else:  # a fraction
+        out = Number(number=str(eval(val_string.replace('?', ''))))
     return out
 
 
@@ -98,17 +107,18 @@ class ViewType(Enum):
 
 class IDP(ASTNode):
     """The class of AST nodes representing an IDP-Z3 program.
-
-    Args:
+    """
+    """ do not display this info in the API
+    Attributes:
         code (str): source code of the IDP program
 
-        vocabularies (dict[str, Vocabulary]): list of vocabulary blocks, by name
+        vocabularies (Dict[str, Vocabulary]): list of vocabulary blocks, by name
 
-        theories (dict[str, Theory]): list of theory blocks, by name
+        theories (Dict[str, TheoryBlock]): list of theory blocks, by name
 
-        structures (dict[str, Structure]): list of structure blocks, by name
+        structures (Dict[str, Structure]): list of structure blocks, by name
 
-        procedures (dict[str, Procedure]): list of procedure blocks, by name
+        procedures (Dict[str, Procedure]): list of procedure blocks, by name
 
         display (Display, Optional): display block, if any
     """
@@ -118,9 +128,12 @@ class IDP(ASTNode):
         self.vocabularies = self.dedup_nodes(kwargs, 'vocabularies')
         self.theories = self.dedup_nodes(kwargs, 'theories')
         self.structures = self.dedup_nodes(kwargs, 'structures')
-        self.display = kwargs.pop('display')
+        displays = kwargs.pop('displays')
         self.procedures = self.dedup_nodes(kwargs, 'procedures')
 
+        assert len(displays) <= 1, "Too many display blocks"
+        self.display = displays[0] if len(displays) == 1 else None
+
         for voc in self.vocabularies.values():
             voc.annotate(self)
         for t in self.theories.values():
@@ -132,7 +145,7 @@ class IDP(ASTNode):
         self.vocabulary = next(iter(self.vocabularies.values()))
         self.theory = next(iter(self.theories    .values()))
         if self.display is None:
-            self.display = Display(constraints=[])
+            self.display = Display(constraints=[], interpretations=[])
 
     @classmethod
     def from_file(cls, file:str) -> "IDP":
@@ -182,14 +195,14 @@ class IDP(ASTNode):
         out.code = code
         return out
 
-    def get_blocks(self, blocks: List[str]):
+    def get_blocks(self, blocks: List[str]) -> List[ASTNode]:
         """returns the AST nodes for the blocks whose names are given
 
         Args:
             blocks (List[str]): list of names of the blocks to retrieve
 
         Returns:
-            List[Union[Vocabulary, Theory, Structure, Procedure, Display]]:
+            List[Union[Vocabulary, TheoryBlock, Structure, Procedure, Display]]:
                 list of AST nodes
         """
         names = blocks.split(",") if type(blocks) is str else blocks
@@ -204,40 +217,13 @@ class IDP(ASTNode):
                        "")
         return out
 
+    def execute(self) -> None:
+        pass  # monkey patched
+
 
 ################################ Vocabulary  ##############################
 
 
-class Annotations(ASTNode):
-    def __init__(self, **kwargs):
-        self.annotations = kwargs.pop('annotations')
-
-        def pair(s):
-            p = s.split(':', 1)
-            if len(p) == 2:
-                try:
-                    # Do we have a Slider?
-                    # The format of p[1] is as follows:
-                    # (lower_sym, upper_sym): (lower_bound, upper_bound)
-                    pat = r"\(((.*?), (.*?))\)"
-                    arg = findall(pat, p[1])
-                    l_symb = arg[0][1]
-                    u_symb = arg[0][2]
-                    l_bound = arg[1][1]
-                    u_bound = arg[1][2]
-                    slider_arg = {'lower_symbol': l_symb,
-                                  'upper_symbol': u_symb,
-                                  'lower_bound': l_bound,
-                                  'upper_bound': u_bound}
-                    return(p[0], slider_arg)
-                except:  # could not parse the slider data
-                    return (p[0], p[1])
-            else:
-                return ('reading', p[0])
-
-        self.annotations = dict((pair(t) for t in self.annotations))
-
-
 class Vocabulary(ASTNode):
     """The class of AST nodes representing a vocabulary block.
     """
@@ -273,27 +259,19 @@ class Vocabulary(ASTNode):
             TypeDeclaration(name=REAL, enumeration=RealRange()),
             TypeDeclaration(name=DATE, enumeration=DateRange()),
             TypeDeclaration(
-                name=SYMBOL,
-                constructors=([Constructor(name=f"`{s}")
-                               for s in [DATE,]]  # TODO '𝔹', 'ℤ', 'ℝ',
-                             +[Constructor(name=f"`{s.name}")
-                                for s in self.declarations
-                                if type(s) == SymbolDeclaration
-                                or type(s) in Type.__args__])),
+                name=CONCEPT,
+                constructors=[]),
+            SymbolDeclaration(annotations='', name=Symbol(name=GOAL_SYMBOL),
+                              sorts=[Subtype(name=CONCEPT, ins=[],
+                                             out=Subtype(name=BOOL))],
+                              out=Subtype(name=BOOL)),
             SymbolDeclaration(annotations='', name=Symbol(name=RELEVANT),
-                                sorts=[], out=Symbol(name=BOOL)),
-            SymbolDeclaration(annotations='', name=Symbol(name=ARITY),
-                                sorts=[Symbol(name=SYMBOL)],
-                                out=Symbol(name=INT)),
+                              sorts=[Subtype(name=CONCEPT, ins=[],
+                                             out=Subtype(name=BOOL))],
+                              out=Subtype(name=BOOL)),
             SymbolDeclaration(annotations='', name=Symbol(name=ABS),
-                                sorts=[Symbol(name=INT)],
-                                out=Symbol(name=INT)),
-            SymbolDeclaration(annotations='', name=Symbol(name=INPUT_DOMAIN),
-                                sorts=[Symbol(name=SYMBOL), Symbol(name=INT)],
-                                out=Symbol(name=SYMBOL)),
-            SymbolDeclaration(annotations='', name=Symbol(name=OUTPUT_DOMAIN),
-                                sorts=[Symbol(name=SYMBOL)],
-                                out=Symbol(name=SYMBOL))
+                                sorts=[Subtype(name=INT)],
+                                out=Subtype(name=INT)),
             ] + self.declarations
 
     def __str__(self):
@@ -305,7 +283,7 @@ class Vocabulary(ASTNode):
         """adds the enumerations in a vocabulary to a theory or structure block
 
         Args:
-            block (Problem): the block to be updated
+            block (Theory): the block to be updated
         """
         for s in self.declarations:
             block.check(s.name not in block.declarations,
@@ -321,12 +299,12 @@ class Vocabulary(ASTNode):
                 block.interpretations[s.name] = s.interpretation
 
 
-class Extern(ASTNode):
+class Import(ASTNode):
     def __init__(self, **kwargs):
         self.name = kwargs.pop('name')
 
     def __str__(self):
-        return f"extern vocabulary {self.name}"
+        return f"Import {self.name}"
 
 
 class TypeDeclaration(ASTNode):
@@ -373,11 +351,19 @@ class TypeDeclaration(ASTNode):
                                  enumeration=enumeration, default=None))
 
     def __str__(self):
-        return (f"type {self.name} := "
-                f"{{{','.join(map(str, self.constructors))}}}")
+        if self.name in RESERVED_SYMBOLS:
+            return ''
+        enumeration = (f"{','.join(map(str, self.constructors))}" if self.constructors else
+                       f"{self.enumeration}")
+        return (f"type {self.name} := {{{enumeration}}}")
 
     def check_bounds(self, var):
-        return self.interpretation.enumeration.contains([var], False)
+        if self.name == CONCEPT:
+            comparisons = [EQUALS([var, UnappliedSymbol.construct(c)])
+                          for c in self.constructors]
+            return OR(comparisons)
+        else:
+            return self.interpretation.enumeration.contains([var], False)
 
     def is_subset_of(self, other):
         return self == other
@@ -401,13 +387,13 @@ class SymbolDeclaration(ASTNode):
 
         arity (int): the number of arguments
 
-        sorts (List[Symbol]): the types of the arguments
+        sorts (List[Subtype]): the types of the arguments
 
-        out (Symbol): the type of the symbol
+        out (Subtype): the type of the symbol
 
         type (string): name of the Z3 type of an instance of the symbol
 
-        domain (List): the list of possible tuples of arguments
+        in_domain (List): the list of possible tuples of arguments
 
         instances (Dict[string, Expression]):
             a mapping from the code of a symbol applied to a tuple of
@@ -422,6 +408,9 @@ class SymbolDeclaration(ASTNode):
 
         heading (str):
             the heading that the symbol should belong to
+
+        optimizable (bool):
+            whether this symbol should get optimize buttons in the IC
     """
 
     def __init__(self, **kwargs):
@@ -445,19 +434,27 @@ class SymbolDeclaration(ASTNode):
         self.private = None
         self.unit: str = None
         self.heading: str = None
+        self.optimizable: bool = True
 
         self.type = None  # a string
-        self.domain = None  # all possible arguments
+        self.in_domain = None  # all possible arguments
         self.range = None  # all possible values
         self.instances = None  # {string: AppliedSymbol} not starting with '_'
         self.block: Optional[Block] = None  # vocabulary where it is declared
         self.view = ViewType.NORMAL  # "hidden" | "normal" | "expanded" whether the symbol box should show atoms that contain that symbol, by default
 
+    @classmethod
+    def make(cls, strname, arity, sorts, out):
+        o = cls(strname=strname, arity=arity, sorts=sorts, out=out, annotations={})
+        return o
+
     def __str__(self):
-        args = ','.join(map(str, self.sorts)) if 0 < len(self.sorts) else ''
-        return (f"{self.name}"
-                f"{ '('+args+')' if args else ''}"
-                f"{'' if self.out.name == BOOL else f' : {self.out.name}'}")
+        if self.name in RESERVED_SYMBOLS:
+            return ''
+        args = '⨯'.join(map(str, self.sorts)) if 0 < len(self.sorts) else ''
+        return (f"{self.name}: "
+                f"{ '('+args+')' if args else '()'}"
+                f" -> {self.out.name}")
 
     def __repr__(self):
         return str(self)
@@ -466,28 +463,41 @@ class SymbolDeclaration(ASTNode):
         return (self.arity == 1 and self.type == BOOL
                 and self.sorts[0].decl == other)
 
-    @classmethod
-    def make(cls, strname, arity, sorts, out):
-        o = cls(strname=strname, arity=arity, sorts=sorts, out=out, annotations={})
-        return o
+    def has_in_domain(self, args: List[Expression]) -> Expression:
+        """Returns an expression that says whether the `args` are in the domain of the symbol.
 
+        Arguments:
+            args (List[Expression]): the list of arguments to be checked, e.g. `[1, 2]`
+
+        Returns:
+            Expression: whether `(1,2)` is in the domain of the symbol
+        """
+        assert len(self.sorts) == len(args), \
+            f"Incorrect arity of {str(args)} for {self.name}"
+        return AND([typ.has_element(term)
+                   for typ, term in zip(self.sorts, args)])
+
+
+    def has_in_range(self, value: Expression) -> Expression:
+        """Returns an expression that says whether `value` is in the range of the symbol.
+        """
+        return self.out.has_element(value)
 
 Type = Union[TypeDeclaration, SymbolDeclaration]
 
 
-################################ Theory  ###############################
+################################ TheoryBlock  ###############################
 
 
-class Theory(ASTNode):
+class TheoryBlock(ASTNode):
     """ The class of AST nodes representing a theory block.
     """
     def __init__(self, **kwargs):
         self.name = kwargs.pop('name')
         self.vocab_name = kwargs.pop('vocab_name')
-        self.constraints = OrderedSet(kwargs.pop('constraints'))
+        constraints = kwargs.pop('constraints')
         self.definitions = kwargs.pop('definitions')
         self.interpretations = self.dedup_nodes(kwargs, 'interpretations')
-        self.goals = {}
 
         self.name = "T" if not self.name else self.name
         self.vocab_name = 'V' if not self.vocab_name else self.vocab_name
@@ -496,8 +506,12 @@ class Theory(ASTNode):
         self.def_constraints = {}  # {(Declaration, Definition): list[Expression]}
         self.assignments = Assignments()
 
-        for constraint in self.constraints:
-            constraint.block = self
+        self.constraints = OrderedSet()
+        for c in constraints:
+            c.block = self
+            if c.annotations is not None:
+                c.expr.annotations = c.annotations.annotations
+            self.constraints.append(c.expr)
         for definition in self.definitions:
             for rule in definition.rules:
                 rule.block = self
@@ -513,25 +527,25 @@ class Definition(ASTNode):
         rules ([Rule]):
             set of rules for the definition, e.g., `!x: p(x) <- q(x)`
 
-        canonicals (dict[Declaration, list[Rule]]):
+        canonicals (Dict[Declaration, list[Rule]]):
             normalized rule for each defined symbol,
             e.g., `!$p!1$: p($p!1$) <- q($p!1$)`
 
-        instantiables (dict[Declaration], list[Expression]):
+        instantiables (Dict[Declaration], list[Expression]):
             list of instantiable expressions for each symbol,
             e.g., `p($p!1$) <=> q($p!1$)`
 
-        clarks (dict[Declaration, Transformed Rule]):
+        clarks (Dict[Declaration, Transformed Rule]):
             normalized rule for each defined symbol (used to be Clark completion)
             e.g., `!$p!1$: p($p!1$) <=> q($p!1$)`
 
-        def_vars (dict[String, dict[String, Variable]]):
+        def_vars (Dict[String, Dict[String, Variable]]):
             Fresh variables for arguments and result
 
-        level_symbols (dict[SymbolDeclaration, Symbol]):
+        level_symbols (Dict[SymbolDeclaration, Symbol]):
             map of recursively defined symbols to level mapping symbols
 
-        cache (dict[SymbolDeclaration, str, Expression]):
+        cache (Dict[SymbolDeclaration, str, Expression]):
             cache of instantiation of the definition
 
         inst_def_level (int): depth of recursion during instantiation
@@ -613,7 +627,7 @@ class Definition(ASTNode):
                 continue
             symbdec = SymbolDeclaration.make(
                 "_"+str(self.id)+"lvl_"+key.name,
-                key.arity, key.sorts, Symbol(name=REAL))
+                key.arity, key.sorts, Subtype(name=REAL))
             self.level_symbols[key] = Symbol(name=symbdec.name)
             self.level_symbols[key].decl = symbdec
 
@@ -658,7 +672,7 @@ class Rule(ASTNode):
 
         Args:
             new_args ([Expression]): tuple of arguments to be applied to the defined symbol
-            theory (Problem): the context for the interpretation
+            theory (Theory): the context for the interpretation
 
         Returns:
             Expression: a boolean expression
@@ -672,7 +686,7 @@ class Rule(ASTNode):
             self.check(len(self.definiendum.sub_exprs) == len(new_args),
                        "Internal error")
             out = out.instantiate(self.definiendum.sub_exprs, new_args, theory)
-            out = AEquivalence.make('⇔', [instance, out])
+            out = EQUIV([instance, out])
         else:
             self.check(len(self.definiendum.sub_exprs) == len(new_args)+1 ,
                        "Internal error")
@@ -699,7 +713,6 @@ class Structure(ASTNode):
         self.name = kwargs.pop('name')
         self.vocab_name = kwargs.pop('vocab_name')
         self.interpretations = self.dedup_nodes(kwargs, 'interpretations')
-        self.goals = {}
 
         self.name = 'S' if not self.name else self.name
         self.vocab_name = 'V' if not self.vocab_name else self.vocab_name
@@ -714,7 +727,7 @@ class Structure(ASTNode):
 
 class SymbolInterpretation(ASTNode):
     """
-    AST node representing `<symbol> := { <identifiers*> } else <default>`
+    AST node representing `<symbol> := { <identifiers*> } else <default>.`
 
     Attributes:
         name (string): name of the symbol being enumerated.
@@ -754,6 +767,7 @@ class SymbolInterpretation(ASTNode):
         else:  # constructs If-then-else recursively
             out = (self.default if self.default is not None else
                    applied._change(sub_exprs=args))
+            tuples.sort(key=lambda t: str(t.args[rank]))
             groups = groupby(tuples, key=lambda t: str(t.args[rank]))
 
             if args[rank].value is not None:
@@ -764,8 +778,8 @@ class SymbolInterpretation(ASTNode):
             else:
                 for val, tuples2 in groups:
                     tuples = list(tuples2)
-                    out = IfExpr.make(
-                        AComparison.make('=', [args[rank], tuples[0].args[rank]]),
+                    out = AIfExpr.make(
+                        EQUALS([args[rank], tuples[0].args[rank]]),
                         self.interpret_application(theory, rank+1,
                                                    applied, args, tuples),
                         out)
@@ -784,7 +798,7 @@ class Enumeration(ASTNode):
     def __init__(self, **kwargs):
         self.tuples = kwargs.pop('tuples')
         if not isinstance(self.tuples, OrderedSet):
-            # self.tuples.sort(key=lambda t: t.code)
+            # self.tuples.sort(key=lambda t: t.code) # do not change dropdown order
             self.tuples = OrderedSet(self.tuples)
         if all(len(c.args) == 1 and type(c.args[0]) == UnappliedSymbol
                for c in self.tuples):
@@ -810,6 +824,7 @@ class Enumeration(ASTNode):
                 "Incorrect arity of tuples in Enumeration.  Please check use of ',' and ';'.")
 
         # constructs If-then-else recursively
+        tuples = sorted(list(tuples), key=lambda t: str(t.args[rank]))
         groups = groupby(tuples, key=lambda t: str(t.args[rank]))
         if args[rank].value is not None:
             for val, tuples2 in groups:  # try to resolve
@@ -818,16 +833,16 @@ class Enumeration(ASTNode):
             return FALSE
         else:
             if rank + 1 == arity:  # use OR
-                out = [ AComparison.make('=', [args[rank], t.args[rank]])
+                out = [ EQUALS([args[rank], t.args[rank]])
                         for t in tuples]
-                out = ADisjunction.make('∨', out)
+                out = OR(out)
                 out.enumerated = ', '.join(str(c) for c in tuples)
                 return out
             out = FALSE
             for val, tuples2 in groups:
                 tuples = list(tuples2)
-                out = IfExpr.make(
-                    AComparison.make('=', [args[rank], tuples[0].args[rank]]),
+                out = AIfExpr.make(
+                    EQUALS([args[rank], tuples[0].args[rank]]),
                     self.contains(args, function, arity, rank+1, tuples),
                     out)
             return out
@@ -846,7 +861,7 @@ class ConstructedFrom(Enumeration):
 
         constructors (List[Constructor]): List of Constructor
 
-        accessors (dict[str, Int]): index of the accessor in the constructors
+        accessors (Dict[str, Int]): index of the accessor in the constructors
     """
     def __init__(self, **kwargs):
         self.constructed = kwargs.pop('constructed')
@@ -857,9 +872,17 @@ class ConstructedFrom(Enumeration):
     def contains(self, args, function, arity=None, rank=0, tuples=None):
         """returns True if args belong to the type enumeration"""
         # args must satisfy the tester of one of the constructors
+        assert len(args) == 1, f"Incorrect arity in {self.parent.name}{args}"
+        if type(args[0].decl) == Constructor:  # try to simplify it
+            self.check(self.parent.name == args[0].decl.type,
+                       f"Incorrect type of {args[0]} for {self.parent.name}")
+            self.check(len(args[0].sub_exprs) == len(args[0].decl.sorts),
+                       f"Incorrect arity")
+            return AND([t.decl.out.has_element(e)
+                        for e,t in zip(args[0].sub_exprs, args[0].decl.sorts)])
         out = [AppliedSymbol.construct(constructor.tester, args)
                 for constructor in self.constructors]
-        return ADisjunction.make('∨', out)
+        return OR(out)
 
 class Tuple(ASTNode):
     def __init__(self, **kwargs):
@@ -923,19 +946,19 @@ class Ranges(Enumeration):
     def contains(self, args, function, arity=None, rank=0, tuples=None):
         var = args[0]
         if not self.elements:
-            return None
+            return TRUE
         if self.tuples and len(self.tuples) < MAX_QUANTIFIER_EXPANSION:
-            es = [AComparison.make('=', [var, c.args[0]]) for c in self.tuples]
-            e = ADisjunction.make('∨', es)
+            es = [EQUALS([var, c.args[0]]) for c in self.tuples]
+            e = OR(es)
             return e
         sub_exprs = []
         for x in self.elements:
             if x.toI is None:
-                e = AComparison.make('=', [var, x.fromI])
+                e = EQUALS([var, x.fromI])
             else:
-                e = AComparison.make(['≤', '≤'], [x.fromI, var, x.toI])
+                e = AComparison.make('≤', [x.fromI, var, x.toI])
             sub_exprs.append(e)
-        return ADisjunction.make('∨', sub_exprs)
+        return OR(sub_exprs)
 
 class IntRange(Ranges):
     def __init__(self):
@@ -957,6 +980,7 @@ class DateRange(Ranges):
 class Display(ASTNode):
     def __init__(self, **kwargs):
         self.constraints = kwargs.pop('constraints')
+        self.interpretations = self.dedup_nodes(kwargs, 'interpretations')
         self.moveSymbols = False
         self.optionalPropagation = False
         self.manualPropagation = False
@@ -965,38 +989,47 @@ class Display(ASTNode):
         self.name = "display"
 
     def run(self, idp):
+        """apply the display block to the idp theory"""
+
+        def base_symbols(name, concepts):
+            """Verify that concepts is a list of concepts.  Returns the list of symbols"""
+            symbols = []
+            # All concepts should be concepts, except for the first
+            # argument of 'unit' and 'heading'.
+            for i, symbol in enumerate(concepts):
+                if name in ['unit', 'heading'] and i == 0:
+                    continue
+                self.check(symbol.name.startswith('`'),
+                    f"arg '{symbol.name}' of {name}'"
+                    f" must begin with a tick '`'")
+                self.check(symbol.name[1:] in self.voc.symbol_decls,
+                    f"argument '{symbol.name}' of '{name}'"
+                    f" must be a concept")
+                symbols.append(self.voc.symbol_decls[symbol.name[1:]])
+            return symbols
+
+        for k, interpretation in self.interpretations.items():
+            symbols = base_symbols(interpretation.name,
+                [t.args[0] for t in interpretation.enumeration.tuples])
+            if interpretation.name == EXPAND:
+                for symbol in symbols:
+                    self.voc.symbol_decls[symbol.name].view = ViewType.EXPANDED
+            elif interpretation.name == GOAL_SYMBOL:
+                idp.theory.interpretations[k] = interpretation
+            else:
+                raise IDPZ3Error(f"Unknown enumeration in display: {interpretation}")
         for constraint in self.constraints:
             if type(constraint) == AppliedSymbol:
                 self.check(type(constraint.symbol.sub_exprs[0]) == Symbol,
                            f"Invalid syntax: {constraint}")
                 name = constraint.symbol.sub_exprs[0].name
-                symbols = []
-                # All arguments should be symbols, except for the first
-                # argument of 'unit' and 'heading'.
-                for i, symbol in enumerate(constraint.sub_exprs):
-                    if name in ['unit', 'heading'] and i == 0:
-                        continue
-                    self.check(symbol.name.startswith('`'),
-                        f"arg '{symbol.name}' of {name}'"
-                        f" must begin with a tick '`'")
-                    self.check(symbol.name[1:] in self.voc.symbol_decls,
-                        f"argument '{symbol.name}' of '{name}'"
-                        f" must be a symbol")
-                    symbols.append(self.voc.symbol_decls[symbol.name[1:]])
+                symbols = base_symbols(name, constraint.sub_exprs)
 
-                if name == 'goal':  # e.g.,  goal(Prime)
-                    for s in symbols:
-                        idp.theory.goals[s.name] = s
-                        s.view = ViewType.EXPANDED  # the goal is always expanded
-                elif name == 'expand':  # e.g. expand(Length, Angle)
-                    for symbol in symbols:
-                        self.voc.symbol_decls[symbol.name].view = ViewType.EXPANDED
-                elif name == 'hide':  # e.g. hide(Length, Angle)
+                if name == 'hide':  # e.g. hide(Length, Angle)
                     for symbol in symbols:
                         self.voc.symbol_decls[symbol.name].view = ViewType.HIDDEN
-                elif name == 'relevant':  # e.g. relevant(Tax)
-                    for s in symbols:
-                        idp.theory.goals[s.name] = s
+                elif name in [GOAL_SYMBOL, EXPAND]:  # e.g. goal_symbol(`tax_amount`)
+                    self.check(False, f"Please use an enumeration for {name}")
                 elif name == 'unit':  # e.g. unit('m', `length):
                     for symbol in symbols:
                         symbol.unit = str(constraint.sub_exprs[0])
@@ -1004,6 +1037,9 @@ class Display(ASTNode):
                     # e.g. heading('Shape', `type).
                     for symbol in symbols:
                         symbol.heading = str(constraint.sub_exprs[0])
+                elif name == 'noOptimization':  # e.g., noOptimization(`temp)
+                    for symbol in symbols:
+                        symbol.optimizable = False
                 elif name == "moveSymbols":
                     self.moveSymbols = True
                 elif name == "optionalPropagation":
@@ -1015,7 +1051,7 @@ class Display(ASTNode):
                 elif name == "manualRelevance":
                     self.manualRelevance = True
                 else:
-                    raise IDPZ3Error(f"unknown display constraint:"
+                    raise IDPZ3Error(f"Unknown display axiom:"
                                      f" {constraint}")
             elif type(constraint) == AComparison:  # e.g. view = normal
                 self.check(constraint.is_assignment(), "Internal error")
@@ -1028,9 +1064,9 @@ class Display(ASTNode):
                                 s.view = ViewType.EXPANDED  # don't change hidden symbols
                     else:
                         self.check(constraint.sub_exprs[1].name == 'normal',
-                                   f"unknown display constraint: {constraint}")
+                                   f"Unknown display axiom: {constraint}")
             else:
-                raise IDPZ3Error(f"unknown display contraint: {constraint}")
+                raise IDPZ3Error(f"Unknown display axiom: {constraint}")
 
 
 ################################ Main  ##################################
@@ -1089,19 +1125,19 @@ class PyAssignment(ASTNode):
 
 ########################################################################
 
-Block = Union[Vocabulary, Theory, Structure, Display]
+Block = Union[Vocabulary, TheoryBlock, Structure, Display]
 
 dslFile = path.join(path.dirname(__file__), 'Idp.tx')
 
 idpparser = metamodel_from_file(dslFile, memoization=True,
                                 classes=[IDP, Annotations,
 
-                                         Vocabulary, Extern,
-                                         TypeDeclaration, Accessor,
+                                         Vocabulary, Import,
+                                         TypeDeclaration, Accessor, Subtype,
                                          SymbolDeclaration, Symbol,
                                          SymbolExpr,
 
-                                         Theory, Definition, Rule, IfExpr,
+                                         TheoryBlock, Definition, Rule, AIfExpr,
                                          AQuantification, Quantee, ARImplication,
                                          AEquivalence, AImplication,
                                          ADisjunction, AConjunction,
