@@ -25,10 +25,11 @@ It has 2 parts:
 * symbolic propagation
 * Z3 propagation
 
-This module monkey-patches the Expression and Problem classes and sub-classes.
+This module monkey-patches the Expression and Theory classes and sub-classes.
 """
 
 import time
+from copy import copy
 from typing import List, Tuple, Optional
 from z3 import (Solver, sat, unsat, unknown, Not, Or, is_false, is_true, is_not, is_eq)
 
@@ -37,11 +38,10 @@ from .Expression import (Expression, AQuantification,
                     ADisjunction, AConjunction, AppliedSymbol,
                     AComparison, AUnary, Brackets, TRUE, FALSE)
 from .Parse import str_to_IDP
-from .Problem import Problem
-from .utils import OrderedSet
+from .Theory import Theory
+from .utils import OrderedSet, IDPZ3Error, NOT_SATISFIABLE
 
 start = time.process_time()
-last_prop = "hello"
 
 ###############################################################################
 #
@@ -61,16 +61,20 @@ def simplify_with(self: Expression, assignments: "Assignments") -> Expression:
     if self.value is not None:
         return self
     value, simpler, new_e, co_constraint = None, None, None, None
-    ass = assignments.get(self.code, None)
-    if ass and ass.value is not None:
-        value = ass.value
     if self.simpler is not None:
         simpler = self.simpler.simplify_with(assignments)
     if self.co_constraint is not None:
         co_constraint = self.co_constraint.simplify_with(assignments)
     new_e = [e.simplify_with(assignments) for e in self.sub_exprs]
-    return self._change(sub_exprs=new_e, value=value, simpler=simpler,
-                        co_constraint=co_constraint).simplify1()
+    self._change(sub_exprs=new_e, simpler=simpler, co_constraint=co_constraint)
+    # calculate ass.value on the changed expression, as simplified sub
+    # expressions may lead to stronger simplifications
+    # E.g., P(C()) where P := {0} and C := 0.
+    ass = assignments.get(self.str, None)
+    if ass and ass.value is not None:
+        value = ass.value
+        self._change(value=value)
+    return self.simplify1()
 Expression.simplify_with = simplify_with
 
 
@@ -175,33 +179,59 @@ Brackets.symbolic_propagate = symbolic_propagate
 #
 ###############################################################################
 
+def _set_consequences_get_changed_choices(self):
+    # clear consequences, as these might not be cleared by add_given when
+    # running via CLI
+    for a in self.assignments.values():
+        if a.status in [S.CONSEQUENCE, S.ENV_CONSQ]:
+            self.assignments.assert__(a.sentence, None, S.UNKNOWN)
 
-def _directional_todo(self):
+    removed_choices = {a.sentence.code: a for a in self.previous_assignments.values()
+                       if a.status in [S.GIVEN, S.DEFAULT, S.EXPANDED]}
+    added_choices = []
+
+    for a in self.assignments.values():
+        if a.status in [S.GIVEN, S.DEFAULT, S.EXPANDED]:
+            if (a.sentence.code in removed_choices
+                    and removed_choices[a.sentence.code].value.same_as(a.value)):
+                removed_choices.pop(a.sentence.code)
+            else:
+                added_choices.append(a)
+
+    if not removed_choices:
+        for a in self.previous_assignments.values():
+            if (a.status in [S.CONSEQUENCE, S.ENV_CONSQ] and
+                    self.assignments[a.sentence.code].status
+                    not in [S.GIVEN, S.EXPANDED, S.DEFAULT]):
+                self.assignments.assert__(a.sentence, a.value, a.status)
+
+    # TODO: why is it not ok to use get_core_atoms in this method?
+
+    return removed_choices, added_choices
+Theory._set_consequences_get_changed_choices = _set_consequences_get_changed_choices
+
+
+def _directional_todo(self, removed_choices={}, added_choices=[]):
     """ computes the list of candidate atoms for a new propagation
-
-    Takes into account assertions made via self.assert_ since the last propagation:
-    * a new assignment forces the re-propagation of Unknowns
-    * a clearing of assignment forces the re-propagation of previous consequences
+    * if choices are removed, all previous consequences and removed choices
+      should be checked for propagation
+    * if choices are added, all unknown atoms should be checked
     """
-    statuses = []
-    if self.propagated:
-        if self.assigned:
-            statuses.extend([S.UNKNOWN, S.EXPANDED])
-        if self.cleared:
-            statuses.extend([S.CONSEQUENCE, S.ENV_CONSQ])
-    else:
-        statuses = [S.UNKNOWN, S.EXPANDED, S.CONSEQUENCE, S.ENV_CONSQ]
 
-    if statuses:
-        todo = OrderedSet(
-            a.sentence for a in self.assignments.values()
-            if ((not a.sentence.is_reified() or self.extended)
-                and a.status in statuses
-                ))
-    else:
-        todo = OrderedSet()
+    todo = {}
+    if removed_choices:
+        for a in removed_choices.values():
+            todo[a.sentence.code] = a.sentence
+        for a in self.previous_assignments.values():
+            if a.status in [S.CONSEQUENCE, S.ENV_CONSQ]:
+                todo[a.sentence.code] = a.sentence
+
+    if added_choices:
+        for a in self.get_core_atoms([S.UNKNOWN]):
+            todo[a.sentence.code] = a.sentence
+
     return todo
-Problem._directional_todo = _directional_todo
+Theory._directional_todo = _directional_todo
 
 
 def _batch_propagate(self, tag=S.CONSEQUENCE):
@@ -225,7 +255,7 @@ def _batch_propagate(self, tag=S.CONSEQUENCE):
                 solver.add(q.reified(self) == q.translate(self))  # in case todo contains complex formula
                 if solver.check() != sat:
                     # print("Falling back !")
-                    yield from self._propagate(tag)
+                    yield from self._propagate(tag=tag)
                 test = Not(q.reified(self) == solver.model().eval(q.reified(self)))  #TODO compute model once
                 tests.append(test)
                 lookup[str(test)] = q
@@ -250,77 +280,174 @@ def _batch_propagate(self, tag=S.CONSEQUENCE):
                     break
                 else:  # unknown
                     # print("Falling back !!")
-                    yield from self._propagate(tag)
+                    yield from self._propagate(tag=tag)
                     break
             yield "No more consequences."
         elif result == unsat:
-            yield "Not satisfiable."
+            yield NOT_SATISFIABLE
             yield str(z3_formula)
         else:
             yield "Unknown satisfiability."
             yield str(z3_formula)
     else:
         yield "No more consequences."
-    self.propagated, self.assigned, self.cleared = True, OrderedSet(), OrderedSet()
-Problem._batch_propagate = _batch_propagate
+Theory._batch_propagate = _batch_propagate
 
 
-def _propagate(self, tag=S.CONSEQUENCE):
-    """generator of new propagated assignments.  Update self.assignments too.
+def _propagate_inner(self, tag, solver, todo):
+    for q in todo.values():
+        solver.add(q.reified(self) == q.translate(self))
+        # reification in case todo contains complex formula
+
+    res1 = solver.check()
+
+    if res1 == sat:
+        model = solver.model()
+        valqs = [(model.eval(q.reified(self)), q) for q in todo.values()]
+        for val1, q in valqs:
+            if str(val1) == str(q.reified(self)):
+                continue  # irrelevant
+            solver.push()
+            solver.add(Not(q.reified(self) == val1))
+            res2 = solver.check()
+            solver.pop()
+
+            assert res2 != unknown, "Incorrect solver behavior"
+            if res2 == unsat:
+                val = str_to_IDP(q, str(val1))
+                yield self.assignments.assert__(q, val, tag)
+
+        yield "No more consequences."
+    elif res1 == unsat:
+        yield NOT_SATISFIABLE
+        yield str(solver.sexpr())
+    else:
+        assert False, "Incorrect solver behavior"
+
+Theory._propagate_inner = _propagate_inner
+
+
+def _first_propagate(self, solver):
+    """ determine universals
+
+        Raises:
+            IDPZ3Error: if theory is unsatisfiable
     """
-    global start, last_prop
-    start, last_prop = time.process_time(), None
-    todo = self._directional_todo()
-    if todo:
-        z3_formula = self.formula()
+    # NOTE: some universal assignments may be set due to the environment theory
+    todo = OrderedSet(a.sentence for a in self.get_core_atoms(
+        [S.UNKNOWN, S.EXPANDED, S.DEFAULT, S.GIVEN, S.CONSEQUENCE, S.ENV_CONSQ]))
 
-        def get_solver():
-            solver = Solver(ctx=self.ctx)
-            solver.add(z3_formula)
-            solver.check()  # required for forall.idp !?
-            for q in todo:
-                solver.add(q.reified(self) == q.translate(self))  # in case todo contains complex formula
-            return solver
+    solver.push()
 
-        solver = get_solver()
-        res1 = solver.check()
-        if res1 == sat:
-            model = solver.model()
-            valqs = [(model.eval(q.reified(self)), q) for q in todo]
-            for val1, q in valqs:
-                if str(val1) == str(q.reified(self)):
-                    continue  # irrelevant
-                solver.push()
-                solver.add(Not(q.reified(self) == val1))
-                res2 = solver.check()
+    for q in todo:
+        solver.add(q.reified(self) == q.translate(self))
+        # reification in case todo contains complex formula
+
+    res1 = solver.check()
+    if res1 == unsat:
+        solver.pop()
+        raise IDPZ3Error(NOT_SATISFIABLE)
+
+    assert res1 == sat, "Incorrect solver behavior"
+    model = solver.model()
+    valqs = [(model.eval(q.reified(self)), q) for q in todo]
+    for val1, q in valqs:
+        if str(val1) == str(q.reified(self)):
+            continue  # irrelevant
+        solver.push()
+        solver.add(Not(q.reified(self) == val1))
+        res2 = solver.check()
+        solver.pop()
+
+        assert res2 != unknown, "Incorrect solver behavior"
+        if res2 == unsat:
+            val = str_to_IDP(q, str(val1))
+
+            ass = self.assignments.get(q.code)
+            if (ass.status in [S.GIVEN, S.DEFAULT, S.EXPANDED] and
+                    not ass.value.same_as(val)):
                 solver.pop()
+                raise IDPZ3Error(NOT_SATISFIABLE)
+            yield self.assignments.assert__(q, val, S.UNIVERSAL)
 
-                if res2 == unsat:
-                    val = str_to_IDP(q, str(val1))
-                    yield self.assignments.assert__(q, val, tag)
-                    last_prop = time.process_time()
-                elif res2 == unknown:  # does not happen with newest version of Z3
-                    solver = get_solver() # restart the solver
-                    solver.check()
-                else:  # reset the value
-                    if self.assignments.get(q, True) is not None:
-                        self.assignments.assert__(q, None, S.UNKNOWN)
-                        last_prop = time.process_time()
-            yield "No more consequences."
-        elif res1 == unsat:
-            yield "Not satisfiable."
-            yield str(z3_formula)
-        else:
-            yield "Unknown satisfiability."
-            yield str(z3_formula)
+    solver.pop()
+Theory._first_propagate = _first_propagate
+
+
+def _propagate_ignored(self, tag=S.CONSEQUENCE, given_todo=None):
+    assert self.ignored_laws, "Internal error"
+
+    solver = self.solver_reified
+    solver.push()
+
+    todo = (given_todo.copy() if given_todo else
+            {a.sentence.code: a.sentence
+            for a in self.assignments.values()
+            if a.status not in [S.GIVEN, S.DEFAULT, S.EXPANDED] and
+            (a.status != S.STRUCTURE or a.sentence.code in self.ignored_laws)})
+
+    self._add_assignment_ignored(solver)
+
+    yield from self._propagate_inner(tag, solver, todo)
+    solver.pop()
+Theory._propagate_ignored = _propagate_ignored
+
+
+def _propagate(self, tag=S.CONSEQUENCE, given_todo=None):
+    """generator of new propagated assignments.  Update self.assignments too.
+    :arg given_todo: custom collection of assignments to check during propagation.
+    given_todo is organized as a dictionary where the keys function to quickly
+    check if a certain assignment is in the collection.
+    """
+
+    global start
+    start = time.process_time()
+
+    if self.ignored_laws:
+        yield from self._propagate_ignored(tag, given_todo)
+        return
+
+    if not self.previous_assignments:
+        try:
+            yield from self._first_propagate(self.solver)
+        except IDPZ3Error:
+            yield NOT_SATISFIABLE
+            # can't access solver.sexpr()
+
+    removed_choices, added_choices = self._set_consequences_get_changed_choices()
+
+    dir_todo = given_todo is None
+    if dir_todo:
+        todo = self._directional_todo(removed_choices, added_choices)
     else:
+        todo = given_todo.copy()  # shallow copy needed because todo might be adjusted
+
+    if not removed_choices and not added_choices:
+        # nothing changed since the previous propagation
+        to_remove = []
+        for a in todo.values():
+            if a.code in self.assignments:
+                to_remove.append(a)
+                if self.assignments[a.code].status in [S.CONSEQUENCE, S.ENV_CONSQ]:
+                    yield self.assignments[a.code]
+        for a in to_remove:
+            todo.pop(a.code)
+
+    if not todo:
         yield "No more consequences."
-    self.propagated, self.assigned, self.cleared = True, OrderedSet(), OrderedSet()
-    if last_prop is None:
-        last_prop = 0
-    else:
-        last_prop -= start
-Problem._propagate = _propagate
+        return
+
+    solver = self.solver
+    solver.push()
+
+    self._add_assignment(solver)
+
+    yield from self._propagate_inner(tag, solver, todo)
+    solver.pop()
+
+    if dir_todo:
+        self.previous_assignments = copy(self.assignments)
+Theory._propagate = _propagate
 
 
 def _z3_propagate(self, tag=S.CONSEQUENCE):
@@ -362,17 +489,16 @@ def _z3_propagate(self, tag=S.CONSEQUENCE):
                 else:
                     print("???", str(consq))
             yield "No more consequences."
-            #yield from self._propagate(tag)  # incomplete --> finish with normal propagation
+            #yield from self._propagate(tag=tag)  # incomplete --> finish with normal propagation
         elif result == unsat:
-            yield "Not satisfiable."
+            yield NOT_SATISFIABLE
             yield str(z3_formula)
         else:
             yield "Unknown satisfiability."
             yield str(z3_formula)
     else:
         yield "No more consequences."
-    self.propagated, self.assigned, self.cleared = True, OrderedSet(), OrderedSet()
-Problem._z3_propagate = _z3_propagate
+Theory._z3_propagate = _z3_propagate
 
 
 Done = True
