@@ -28,14 +28,14 @@ import sys
 from typing import List
 
 from .Expression import (
-    Constructor, Expression, IfExpr, AQuantification, Quantee,
+    Constructor, Expression, AIfExpr, AQuantification, Quantee, Subtype,
     Operator, AEquivalence, AImplication, ADisjunction,
     AConjunction, AComparison, ASumMinus, AMultDiv, APower,
-    AUnary, AAggregate, SymbolExpr, AppliedSymbol, UnappliedSymbol,
-    Number, Date, Brackets, TRUE, FALSE)
-from .Parse import Symbol, Enumeration, Tuple
+    AUnary, AAggregate, SymbolExpr, AppliedSymbol, UnappliedSymbol, Variable,
+    Number, Date, Brackets, TRUE, FALSE, NOT, AND, OR)
+from .Parse import Symbol, Enumeration, Tuple, TypeDeclaration
 from .Assignments import Status as S, Assignment
-from .utils import BOOL, INT, SYMBOL, ABS, ARITY, INPUT_DOMAIN, OUTPUT_DOMAIN
+from .utils import BOOL, INT, CONCEPT, ABS, RESERVED_SYMBOLS
 
 
 # class Expression  ###########################################################
@@ -59,7 +59,7 @@ def _change(self, sub_exprs=None, ops=None, value=None, simpler=None,
             self.simpler = simpler
     assert (self.value is None
                or type(self.value) in [AppliedSymbol, UnappliedSymbol, Symbol,
-                                       Number, Date]), \
+                                       Number, Date, Subtype]), \
             f"Incorrect value in _change: {self.value}"
 
     # reset derived attributes
@@ -82,7 +82,7 @@ Expression.simplify1 = simplify1
 
 
 
-# Class IfExpr  ###############################################################
+# Class AIfExpr  ###############################################################
 
 
 def update_exprs(self, new_exprs):
@@ -98,10 +98,10 @@ def update_exprs(self, new_exprs):
         elif then_.same_as(TRUE) and else_.same_as(FALSE):
             return self._change(simpler=if_, sub_exprs=sub_exprs)
         elif then_.same_as(FALSE) and else_.same_as(TRUE):
-            return self._change(simpler=AUnary.make('¬', if_),
+            return self._change(simpler=NOT(if_),
                                 sub_exprs=sub_exprs)
     return self._change(sub_exprs=sub_exprs)
-IfExpr.update_exprs = update_exprs
+AIfExpr.update_exprs = update_exprs
 
 
 # Class Quantee  #######################################################
@@ -122,9 +122,9 @@ def update_exprs(self, new_exprs):
     exprs = list(new_exprs)
     if not self.quantees:
         if self.q == '∀':
-            simpler = AConjunction.make('∧', exprs)
+            simpler = AND(exprs)
         else:
-            simpler = ADisjunction.make('∨', exprs)
+            simpler = OR(exprs)
         return self._change(simpler=simpler, sub_exprs=[simpler])
     return self._change(sub_exprs=exprs)
 AQuantification.update_exprs = update_exprs
@@ -151,7 +151,7 @@ def update_exprs(self, new_exprs):
             exprs0 = exprs0 if self.sub_exprs[0].same_as(TRUE) else TRUE
             value = TRUE
         elif exprs1.same_as(FALSE):  # (p => false) is ~p
-            simpler = AUnary.make('¬', exprs0)
+            simpler = NOT(exprs0)
     return self._change(value=value, simpler=simpler,
                         sub_exprs=[exprs0, exprs1])
 AImplication.update_exprs = update_exprs
@@ -165,10 +165,10 @@ def update_exprs(self, new_exprs):
         return self._change(simpler=exprs[1], sub_exprs=exprs)
     for e in exprs:
         if e.same_as(TRUE):  # they must all be true
-            return self._change(simpler=AConjunction.make('∧', exprs),
+            return self._change(simpler=AND(exprs),
                                 sub_exprs=exprs)
         if e.same_as(FALSE):  # they must all be false
-            return self._change(simpler=AConjunction.make('∧', [AUnary.make('¬', e) for e in exprs]),
+            return self._change(simpler=AND([NOT(e) for e in exprs]),
                                 sub_exprs=exprs)
     return self._change(sub_exprs=exprs)
 AEquivalence.update_exprs = update_exprs
@@ -252,9 +252,12 @@ def update_arith(self, family, operands):
     operands = list(operands)
     operands1 = [e.value for e in operands]
     if all(e is not None for e in operands1):
+        self.check(all(hasattr(e, 'py_value') for e in operands1),
+                f"Incorrect numeric type in {self}")
         out = operands1[0].py_value
 
-        for e, op in zip(operands1[1:], self.operator):
+        assert len(self.operator) == len(operands1[1:]), "Internal error"
+        for op, e in zip(self.operator, operands1[1:]):
             function = Operator.MAP[op]
 
             if op == '/' and self.type == INT:  # integer division
@@ -332,7 +335,7 @@ AUnary.as_set_condition = as_set_condition
 
 def update_exprs(self, new_exprs):
     operands = list(new_exprs)
-    if self.using_if and not self.quantees:
+    if self.annotated and not self.quantees:
         operands1 = [e.value for e in operands]
         if all(e is not None for e in operands1):
             out = sum(e.py_value for e in operands1)
@@ -374,39 +377,6 @@ def update_exprs(self, new_exprs):
         if self.decl.name == new_exprs[0].decl.tester.name:
             return self._change(value=TRUE, sub_exprs=new_exprs)
 
-    # simplify arity(s), input_domain(s,0), output_domain(s)
-    if (self.decl and new_exprs
-        and type(new_exprs[0]) == UnappliedSymbol and new_exprs[0].decl):
-        if self.decl.name == ARITY:
-            self.check(len(new_exprs) == 1,
-                    f"Incorrect number of arguments for '{ARITY}': {len(new_exprs)}")
-            self.check(new_exprs[0].decl.type == SYMBOL,
-                    f"Argument of '{ARITY}' must be a Symbol: {new_exprs[0]}")
-            value = Number(number=str(new_exprs[0].decl.symbol.decl.arity))
-            return self._change(value=value, sub_exprs=new_exprs)
-        elif self.decl.name == INPUT_DOMAIN:
-            self.check(len(new_exprs) == 2,
-                    f"Incorrect number of arguments for '{INPUT_DOMAIN}': {len(new_exprs)}")
-            self.check(new_exprs[0].decl.type == SYMBOL
-                       or (new_exprs[0].decl.sort.decl.arity == 1
-                           and new_exprs[0].decl.sort.decl.type == BOOL),
-                    f"First argument of '{INPUT_DOMAIN}' must be a Symbol: {new_exprs[0]}")
-            self.check(new_exprs[1].type == INT,
-                    f"Second argument of '{INPUT_DOMAIN}' must be a Int: {new_exprs[1]}")
-            if isinstance(new_exprs[1], Number):
-                # find the Symbol for the input domain
-                symbol_string = f"`{new_exprs[0].decl.symbol.decl.sorts[new_exprs[1].py_value - 1]}"
-                value = self.decl.out.decl.map[symbol_string]
-                return self._change(value=value, sub_exprs=new_exprs)
-        elif self.decl.name == OUTPUT_DOMAIN:
-            self.check(len(new_exprs) == 1,
-                    f"Incorrect number of arguments for '{OUTPUT_DOMAIN}': {len(new_exprs)}")
-            self.check(new_exprs[0].decl.type == SYMBOL,
-                    f"Argument of '{OUTPUT_DOMAIN}' must be a Symbol: {new_exprs[0]}")
-            # find the Symbol for the output domain of the argument
-            symbol_string = f"`{new_exprs[0].decl.symbol.decl.out}"
-            value = self.decl.out.decl.map[symbol_string]
-            return self._change(value=value, sub_exprs=new_exprs)
     return self._change(sub_exprs=new_exprs)
 AppliedSymbol.update_exprs = update_exprs
 
@@ -426,6 +396,8 @@ def update_exprs(self, new_exprs):
     value = (symbol if self.eval == '' else
              symbol.decl.symbol if type(symbol) == UnappliedSymbol and symbol.decl else
              None)
+    self.check(not value or type(value) != Variable,
+               f"Variable `{value}` cannot be applied to argument(s).")
     self.decl = value.decl if value else None
     return self._change(sub_exprs=[symbol], value=value)
 SymbolExpr.update_exprs = update_exprs
